#!/usr/bin/env python3
"""DFT functions that are only needed for (meta-)GGA calculations.

Most functions here have been optimized for performance and can be harder to understand than normal.
To mitigate this, easier (but slower) implementations have been added as comments.
"""
import numpy as np


def get_grad_field(atoms, field, real=True):
    """Calculate the gradient of fields on the grid per spin channel.

    Args:
        atoms: Atoms object.
        field (ndarray): Real-space field per spin channel.

    Keyword Args:
        real (bool): Make the gradient a real array (the gradient of n_spin is real).

    Returns:
        ndarray: Gradients of field per spin channel.
    """
    dfield = np.empty((atoms.occ.Nspin, atoms.Ns, 3), dtype=complex)
    for spin in range(atoms.occ.Nspin):
        fieldG = atoms.J(field[spin])
        for dim in range(3):
            dfield[spin, :, dim] = atoms.I(1j * atoms.G[:, dim] * fieldG)
    if real:
        return np.real(dfield)
    return dfield


def gradient_correction(atoms, spin, dn_spin, vsigma):
    """Calculate the gradient correction for the exchange-correlation potential.

    Reference: Chem. Phys. Lett. 199, 557.

    Args:
        atoms: Atoms object.
        spin (int): Spin variable to track weather to do the calculation for spin up or down.
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        vsigma (ndarray): Contracted gradient potential derivative.

    Returns:
        ndarray: Gradient correction in reciprocal space.
    """
    # sigma is |dn|^2, while vsigma is n * d exc/d sigma
    h = np.zeros_like(dn_spin)
    if not atoms.unrestricted:
        # In the unpolarized case we have no spin mixing and only one spin density
        h[0] = 2 * vsigma[0, :, None] * dn_spin[0]
    else:
        # In the polarized case we would get for spin up (and similar for spin down)
        # Vxc_u = vxc_u - Nabla dot (2 vsigma_uu * dn_u + vsigma_ud * dn_d)
        # h is the expression in the brackets
        h[0] = 2 * vsigma[0, :, None] * dn_spin[0] + vsigma[1, :, None] * dn_spin[1]
        h[1] = 2 * vsigma[2, :, None] * dn_spin[1] + vsigma[1, :, None] * dn_spin[0]

    # Calculate Nabla dot h
    # Normally we would calculate the correction for each spin, but we only need one at a time in H
    Gh = np.empty((len(atoms.G2), 3), dtype=complex)
    for dim in range(3):
        Gh[:, dim] = atoms.J(h[spin, :, dim])
    # return 1j * np.sum(atoms.G * Gh, axis=1)
    return 1j * np.einsum('ir,ir->i', atoms.G, Gh)


def get_tau(atoms, Y):
    """Calculate the positive-definite kinetic energy densities per spin.

    Reference: J. Chem. Phys. 109, 2092.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        ndarray: Real-space positive-definite kinetic energy density.
    """
    # The "intuitive" way is the one commented out below
    # Sadly, this implementation is really slow for various reasons so use the faster one below

    # Yrs = atoms.I(Y)
    # tau = np.zeros((atoms.occ.Nspin, atoms.Ns), dtype=complex)
    # for i in range(atoms.occ.Nstate):
    #     dYrs = get_grad_field(atoms, Yrs[..., i], real=False)
    #     tau += 0.5 * atoms.occ.f[:, i, None] * np.sum(dYrs.conj() * dYrs, axis=2)
    # return np.real(tau)

    dYrs = np.empty((atoms.occ.Nspin, atoms.Ns, atoms.occ.Nstate, 3), dtype=complex)
    # Calculate the active G vectors and broadcast to a desired shape
    Gc = atoms.G[atoms.active][None, :, None, :]
    # Calculate the gradients of Y in the active(!) reciprocal space and transform to real-space
    for dim in range(3):
        dYrs[..., dim] = atoms.I(1j * Gc[..., dim] * Y)
    # Sum over dimensions (dYx* dYx + dYy* dYy + dYz* dYz)
    # sumdYrs = np.real(np.sum(dYrs.conj() * dYrs, axis=3))
    # Sum over all states
    # Use the definition with a division by two
    # return 0.5 * np.sum(atoms.occ.f[:, None, :] * sumdYrs, axis=2)
    # Or in compressed Einstein notation:
    return 0.5 * np.real(np.einsum('sj,sijr,sijr->si', atoms.occ.f, dYrs.conj(), dYrs))


def calc_Vtau(scf, spin, W, vtau):
    """Calculate the tau-dependent potential contribution for meta-GGAs.

    Reference: J. Chem. Phys. 145, 204114.

    Args:
        scf: SCF object.
        spin (int): Spin variable to track weather to do the calculation for spin up or down.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        vtau (ndarray): Kinetic energy density potential derivative.

    Returns:
        ndarray: Tau-dependent potential contribution in reciprocal space.
    """
    atoms = scf.atoms

    Vpsi = np.zeros((len(atoms.G2c), atoms.occ.Nstate), dtype=complex)
    if scf.xc_type == 'meta-gga':  # Only calculate the contribution for meta-GGAs
        # The "intuitive" way is the one commented out below
        # Sadly, this implementation is really slow for various reasons so use the faster one below

        # GVpsi = np.empty((len(atoms.G2c), 3), dtype=complex)
        # Gc = atoms.G[atoms.active]
        # Wrs = atoms.I(W)
        # for i in range(atoms.occ.Nstate):
        #     dWrs = get_grad_field(atoms, Wrs[..., i], real=False)
        #     for r in range(3):
        #         GVpsi[:, r] = atoms.J(vtau[spin] * dWrs[spin, :, r], full=False)
        #     Vpsi[:, i] = -0.5 * 1j * np.sum(Gc * GVpsi, axis=1)

        GVpsi = np.empty((len(atoms.G2c), atoms.occ.Nstate, 3), dtype=complex)
        dWrs = np.empty((atoms.Ns, atoms.occ.Nstate, 3), dtype=complex)
        # Calculate the active G vectors and broadcast to a desired shape
        Gc = atoms.G[atoms.active][:, None, :]
        # We only calculate Vtau for one spin channel, index, and reshape before the loop
        vtau_spin = vtau[spin, :, None]
        W_spin = W[spin]
        for dim in range(3):
            # Calculate the gradients of W in the active(!) space and transform to real-space
            dWrs[..., dim] = atoms.I(1j * Gc[..., dim] * W_spin)
            # Calculate dexc/dtau * gradpsi and transform to the active reciprocal space
            GVpsi[..., dim] = atoms.J(vtau_spin * dWrs[..., dim], full=False)
        # Sum over dimensions
        # Calculate -0.5 Nabla dot Gvpsi (compare with gradient_correction)
        # Vpsi = -0.5 * 1j * np.sum(Gc * GVpsi, axis=2)
        Vpsi = -0.5 * 1j * np.einsum('ior,ijr->ij', Gc, GVpsi)
    return Vpsi * atoms.Omega



#!/usr/bin/env python3
"""Main DFT functions based on the DFT++ formulation."""
import numpy as np
from numpy.random import Generator, SFC64
from scipy.linalg import eig, eigh, eigvalsh, inv, sqrtm

from .gga import calc_Vtau, get_grad_field, get_tau, gradient_correction
from .gth import calc_Vnonloc
from .utils import handle_spin_gracefully, pseudo_uniform
from .xc import get_vxc


def solve_poisson(atoms, n):
    """Solve the Poisson equation.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        n (ndarray): Real-space electronic density.

    Returns:
        ndarray: Hartree field.
    """
    # phi = -4 pi Linv(O(J(n)))
    return -4 * np.pi * atoms.Linv(atoms.O(atoms.J(n)))


def get_n_total(atoms, Y, n_spin=None):
    """Calculate the total electronic density.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Keyword Args:
        n_spin (ndarray): Real-space electronic densities per spin channel.

    Returns:
        ndarray: Electronic density.
    """
    # Return the total density in the spin-paired case
    if n_spin is not None:
        return np.sum(n_spin, axis=0)

    # n = (IW) F (IW)dag
    Yrs = atoms.I(Y)
    n = np.zeros(atoms.Ns)
    for spin in range(atoms.occ.Nspin):
        n += np.sum(atoms.occ.f[spin] * np.real(Yrs[spin].conj() * Yrs[spin]), axis=1)
    return n


def get_n_spin(atoms, Y):
    """Calculate the electronic density per spin channel.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        ndarray: Electronic densities per spin channel.
    """
    Yrs = atoms.I(Y)
    n = np.empty((atoms.occ.Nspin, atoms.Ns))
    for spin in range(atoms.occ.Nspin):
        n[spin] = np.sum(atoms.occ.f[spin] * np.real(Yrs[spin].conj() * Yrs[spin]), axis=1)
    return n


def get_n_single(atoms, Y):
    """Calculate the single-electron densities.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        ndarray: Single-electron densities.
    """
    Yrs = atoms.I(Y)
    n = np.empty((atoms.occ.Nspin, atoms.Ns, atoms.occ.Nstate))
    for spin in range(atoms.occ.Nspin):
        n[spin] = atoms.occ.f[spin] * np.real(Yrs[spin].conj() * Yrs[spin])
    return n


@handle_spin_gracefully
def orth(atoms, W):
    """Orthogonalize coefficient matrix W.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: Orthogonalized wave functions.
    """
    # Y = W (Wdag O(W))^-0.5
    return W @ inv(sqrtm(W.conj().T @ atoms.O(W)))


def get_grad(scf, spin, W, **kwargs):
    """Calculate the energy gradient with respect to W.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        spin (int): Spin variable to track weather to do the calculation for spin up or down.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        **kwargs: See :func:`H`.

    Returns:
        ndarray: Gradient.
    """
    atoms = scf.atoms
    F = atoms.occ.F[spin]
    HW = H(scf, spin, W, **kwargs)
    WHW = W[spin].conj().T @ HW
    # U = Wdag O(W)
    OW = atoms.O(W[spin])
    U = W[spin].conj().T @ OW
    invU = inv(U)
    U12 = sqrtm(invU)
    # Htilde = U^-0.5 Wdag H(W) U^-0.5
    Ht = U12 @ WHW @ U12
    # grad E = H(W) - O(W) U^-1 (Wdag H(W)) (U^-0.5 F U^-0.5) + O(W) (U^-0.5 Q(Htilde F - F Htilde))
    return (HW - (OW @ invU) @ WHW) @ (U12 @ F @ U12) + OW @ (U12 @ Q(Ht @ F - F @ Ht, U))


def H(scf, spin, W, dn_spin=None, phi=None, vxc=None, vsigma=None, vtau=None):
    """Left-hand side of the eigenvalue equation.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        spin (int): Spin variable to track weather to do the calculation for spin up or down.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        phi (ndarray): Hartree field.
        vxc (ndarray): Exchange-correlation potential.
        vsigma (ndarray): Contracted gradient potential derivative.
        vtau (ndarray): Kinetic energy gradient potential derivative.

    Returns:
        ndarray: Hamiltonian applied on W.
    """
    atoms = scf.atoms

    # If dn_spin is None all other keyword arguments are None by design
    # In that case precompute values from the SCF class
    if dn_spin is None:
        dn_spin, phi, vxc, vsigma, vtau = H_precompute(scf, W)

    # This calculates the XC potential in the reciprocal space
    Gvxc = atoms.J(vxc[spin])
    # Calculate the gradient correction to the potential if a (meta-)GGA functional is used
    if 'gga' in scf.xc_type:
        Gvxc = Gvxc - gradient_correction(atoms, spin, dn_spin, vsigma)
    # Vkin = -0.5 L(W)
    Vkin_psi = -0.5 * atoms.L(W[spin])
    # Veff = Jdag(Vion) + Jdag(O(J(vxc))) + Jdag(O(phi))
    # We get the full potential in the functional definition (different to the DFT++ notation)
    # Normally Vxc = Jdag(O(J(exc))) + diag(exc') Jdag(O(J(n))) (for LDA functionals)
    Veff = scf.Vloc + atoms.Jdag(atoms.O(Gvxc + phi))
    Vnonloc_psi = calc_Vnonloc(scf, spin, W)
    Vtau_psi = calc_Vtau(scf, spin, W, vtau)
    # H = Vkin + Idag(diag(Veff))I + Vnonloc (+ Vtau)
    # Diag(a) * B can be written as a * B if a is a column vector
    return Vkin_psi + atoms.Idag(Veff[:, None] * atoms.I(W[spin])) + Vnonloc_psi + Vtau_psi


def H_precompute(scf, W):
    """Create precomputed values as intermediate results.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        tuple[ndarray, ndarray, ndarray, ndarray, ndarray]: dn_spin, phi, vxc, vsigma, and vtau
    """
    # We are calculating everything here over both spin channels
    # We would be fine/could improve performance by only calculating the currently used spin channel
    atoms = scf.atoms
    Y = orth(atoms, W)
    n_spin = get_n_spin(atoms, Y)
    n = get_n_total(atoms, Y, n_spin)
    if 'gga' in scf.xc_type:
        dn_spin = get_grad_field(atoms, n_spin)
    else:
        dn_spin = None
    if scf.xc_type == 'meta-gga':
        tau = get_tau(atoms, Y)
    else:
        tau = None
    phi = solve_poisson(atoms, n)
    vxc, vsigma, vtau = get_vxc(scf.xc, n_spin, atoms.occ.Nspin, dn_spin, tau)
    return dn_spin, phi, vxc, vsigma, vtau


def Q(inp, U):
    """Operator needed to calculate gradients with non-constant occupations.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        inp (ndarray): Coefficients input array.
        U (ndarray): Overlap of wave functions.

    Returns:
        ndarray: Q operator result.
    """
    mu, V = eig(U)
    mu = mu[:, None]
    denom = np.sqrt(mu) @ np.ones((1, len(mu)))
    denom2 = denom + denom.conj().T
    return V @ ((V.conj().T @ inp @ V) / denom2) @ V.conj().T


def get_psi(scf, W):
    """Calculate eigenstates from H.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: Eigenstates in reciprocal space.
    """
    atoms = scf.atoms
    Y = orth(atoms, W)
    psi = np.empty_like(Y)
    for spin in range(atoms.occ.Nspin):
        mu = Y[spin].conj().T @ H(scf, spin, Y)
        _, D = eigh(mu)
        psi[spin] = Y[spin] @ D
    return psi


def get_epsilon(scf, W):
    """Calculate eigenvalues from H.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: Eigenvalues.
    """
    atoms = scf.atoms
    Y = orth(atoms, W)
    epsilon = np.empty((atoms.occ.Nspin, atoms.occ.Nstate))
    for spin in range(atoms.occ.Nspin):
        mu = Y[spin].conj().T @ H(scf, spin, Y)
        epsilon[spin] = np.sort(eigvalsh(mu))
    return epsilon


def guess_random(scf, seed=42, symmetric=False):
    """Generate random initial-guess coefficients as starting values.

    Args:
        scf: SCF object.

    Keyword Args:
        seed (int): Seed to initialize the random number generator.
        symmetric (bool): Weather to use the same guess for both spin channels.

    Returns:
        ndarray: Initial-guess orthogonal wave functions in reciprocal space.
    """
    atoms = scf.atoms
    rng = Generator(SFC64(seed))
    if symmetric:
        W = rng.standard_normal((len(atoms.G2c), atoms.occ.Nstate)) + \
            1j * rng.standard_normal((len(atoms.G2c), atoms.occ.Nstate))
        W = np.array([W] * atoms.occ.Nspin)
    else:
        W = rng.standard_normal((atoms.occ.Nspin, len(atoms.G2c), atoms.occ.Nstate)) + \
            1j * rng.standard_normal((atoms.occ.Nspin, len(atoms.G2c), atoms.occ.Nstate))
    return orth(atoms, W)


def guess_pseudo(scf, seed=1234, symmetric=False):
    """Generate initial-guess coefficients using pseudo-random starting values.

    Args:
        scf: SCF object.

    Keyword Args:
        seed (int): Seed to initialize the random number generator.
        symmetric (bool): Weather to use the same guess for both spin channels.

    Returns:
        ndarray: Initial-guess orthogonal wave functions in reciprocal space.
    """
    atoms = scf.atoms
    if symmetric:
        W = pseudo_uniform((1, len(atoms.G2c), atoms.occ.Nstate), seed=seed)
        W = np.array([W[0]] * atoms.occ.Nspin)
    else:
        W = pseudo_uniform((atoms.occ.Nspin, len(atoms.G2c), atoms.occ.Nstate), seed=seed)
    return orth(atoms, W)



#!/usr/bin/env python3
"""Modified Perdew-Wang LDA correlation.

Reference: Phys. Rev. B 45, 13244.
"""
from .lda_c_pw import lda_c_pw, lda_c_pw_spin


def lda_c_pw_mod(n, **kwargs):
    """Modified Perdew-Wang parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_PW_MOD and ID 13 in Libxc.

    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    """
    return lda_c_pw(n, A=0.0310907, **kwargs)


def lda_c_pw_mod_spin(n, zeta, **kwargs):
    """Modified Perdew-Wang parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_PW_MOD and ID 13 in Libxc.

    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    """
    # fzeta0 = 4 / (9 * (2**(1 / 3) - 1))
    return lda_c_pw_spin(n, zeta, A=(0.0310907, 0.01554535, 0.0168869),
                         fzeta0=1.709920934161365617563962776245, **kwargs)



#!/usr/bin/env python3
"""Vosko-Wilk-Nusair LDA correlation.

Reference: Phys. Rev. B 22, 3812.
"""
import numpy as np


def lda_c_vwn(n, A=0.0310907, b=3.72744, c=12.9352, x0=-0.10498, **kwargs):
    """Vosko-Wilk-Nusair parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_VWN and ID 7 in Libxc.

    Reference: Phys. Rev. B 22, 3812.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        A (float): Functional parameter.
        b (float): Functional parameter.
        c (float): Functional parameter.
        x0 (float): Functional parameter.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: VWN correlation energy density and potential.
    """
    rs = (3 / (4 * np.pi * n))**(1 / 3)

    x = np.sqrt(rs)
    X = rs + b * x + c
    Q = np.sqrt(4 * c - b**2)
    fx0 = b * x0 / (x0**2 + b * x0 + c)
    f3 = 2 * (2 * x0 + b) / Q
    tx = 2 * x + b
    tanx = np.arctan(Q / tx)

    ec = A * (np.log(rs / X) + 2 * b / Q * tanx - fx0 * (np.log((x - x0)**2 / X) + f3 * tanx))

    tt = tx**2 + Q**2
    vc = ec - x * A / 6 * (2 / x - tx / X - 4 * b / tt -
                           fx0 * (2 / (x - x0) - tx / X - 4 * (2 * x0 + b) / tt))
    return ec, np.array([vc]), None


def lda_c_vwn_spin(n, zeta, **kwargs):
    """Vosko-Wilk-Nusair parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_VWN and ID 7 in Libxc.

    Reference: Phys. Rev. B 22, 3812.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: VWN correlation energy density and potential.
    """
    A = (0.0310907, 0.01554535, -1 / (6 * np.pi**2))
    b = (3.72744, 7.06042, 1.13107)
    c = (12.9352, 18.0578, 13.0045)
    x0 = (-0.10498, -0.325, -0.0047584)

    ec0, vc0, _ = lda_c_vwn(n, A[0], b[0], c[0], x0[0])  # Paramagnetic
    ec1, vc1, _ = lda_c_vwn(n, A[1], b[1], c[1], x0[1])  # Ferromagnetic
    ac, dac, _ = lda_c_vwn(n, A[2], b[2], c[2], x0[2])   # Spin stiffness

    d2fzeta0 = 4 / 9 / (2**(1 / 3) - 1)
    fzeta = 0.5 * ((1 + zeta)**(4 / 3) + (1 - zeta)**(4 / 3) - 2) / (2**(1 / 3) - 1)
    zeta4 = zeta**4
    fzetaz4 = fzeta * zeta4
    De = ec1 - ec0 - ac / d2fzeta0

    ec = ec0 + ac / d2fzeta0 * fzeta + De * fzetaz4

    dfzeta = 4 / 6 * ((1 + zeta)**(1 / 3) - (1 - zeta)**(1 / 3)) / (2**(1 / 3) - 1)
    dec1 = vc0 + dac / d2fzeta0 * fzeta + (vc1 - vc0 - dac / d2fzeta0) * fzetaz4
    dec2 = ac / d2fzeta0 * dfzeta + De * (4 * zeta**3 * fzeta + zeta4 * dfzeta)

    vcup = dec1 + (1 - zeta) * dec2
    vcdw = dec1 - (1 + zeta) * dec2
    return ec, np.array([vcup, vcdw]), None



#!/usr/bin/env python3
"""Logger initialization and configuration."""
import logging
import numbers
import sys


class CustomLogger(logging.Logger):
    """Custom logger for the usage outside of classes.

    This is just a basic logger but with an added verbose property.

    Args:
        name (str): Logger name.
    """
    def __init__(self, name):
        """Initialize the CustomLogger object."""
        super().__init__(name)

    @property
    def verbose(self):
        """Verbosity level."""
        return self._verbose

    @verbose.setter
    def verbose(self, level):
        self._verbose = get_level(level)
        self.setLevel(self._verbose)


class CustomFormatter(logging.Formatter):
    """Custom logger formatter."""
    def format(self, record):  # noqa: A003
        """Use different formatting for different logging levels.

        Args:
            record: LogRecord object.
        """
        if record.levelno >= logging.WARNING:
            # Print the level name for errors and warnings
            self._style._fmt = '%(levelname)s: %(msg)s'
        else:
            # But not for info and debug messages
            self._style._fmt = '%(msg)s'
        return super().format(record)


# The following code is not guarded by a function because it has to be run once the logger is called
# to set up the basic logger configuration

# Create a base logger that can be used outside of classes
logging.setLoggerClass(CustomLogger)
#: Global logging object.
log = logging.getLogger('eminus')

# Basic logger setup
__formatter = CustomFormatter()
__handler = logging.StreamHandler(sys.stdout)
__handler.setFormatter(__formatter)
logging.root.addHandler(__handler)


def create_logger(obj):
    """Create a logger unique to an object.

    Args:
        obj: Instance of a class.
    """
    # Use the ID of objects to create a unique logger
    # Without this setting the verbosity in one instance would affect other instances
    local_log = logging.getLogger(str(id(obj)))
    local_log.verbose = log.verbose
    return local_log


def get_level(verbose):
    """Validate logging levels.

    Args:
        verbose (int | str): Level of output.
    """
    log_levels = {
        0: 'CRITICAL',
        1: 'ERROR',
        2: 'WARNING',
        3: 'INFO',
        4: 'DEBUG'
    }
    # Use the global logging level for None
    if verbose is None:
        level = log.verbose
    # Fall back to DEBUG if the level is not available
    elif isinstance(verbose, numbers.Number):
        level = log_levels.get(verbose, 'DEBUG')
    else:
        level = verbose
    level = level.upper()
    if level not in log_levels.values():
        ValueError(f'{level} is no recognized logging level.')
    return level


def name(newname):
    """Add a name to functions without evaluating them for better logging.

    Args:
        newname (str): Function name.

    Returns:
        Callable: Decorator.
    """
    def decorator(f):
        f.__name__ = newname
        return f
    return decorator



#!/usr/bin/env python3
"""Functions to restrict real-space fields to domains."""
import numbers

import numpy as np
from scipy.linalg import norm

from .tools import center_of_mass


def domain_cuboid(obj, length, centers=None):
    """Generate a mask for a cuboidal real-space domain.

    Args:
        obj: Atoms or SCF object.
        length (int | float | list | ndarray): Side length or lengths of the cuboid.

    Keyword Args:
        centers (ndarray): Center of the cuboid.

            Defaults to the geometric center of mass of the system. For multiple coordinates,
            multiple domains will be merged.

    Returns:
        ndarray: Boolean mask.
    """
    atoms = obj._atoms

    if isinstance(length, numbers.Real):
        length = length * np.ones(3)
    if centers is None:
        centers = center_of_mass(atoms.X)
    centers = np.asarray(centers)
    # Handle each dimension separately and add them together
    if centers.ndim == 1:
        mask1 = np.abs(centers[0] - atoms.r[:, 0]) < length[0]
        mask2 = np.abs(centers[1] - atoms.r[:, 1]) < length[1]
        mask3 = np.abs(centers[2] - atoms.r[:, 2]) < length[2]
        mask = mask1 & mask2 & mask3
    else:
        mask = np.zeros(atoms.Ns, dtype=bool)
        for center in centers:
            mask1 = np.abs(center[0] - atoms.r[:, 0]) < length[0]
            mask2 = np.abs(center[1] - atoms.r[:, 1]) < length[1]
            mask3 = np.abs(center[2] - atoms.r[:, 2]) < length[2]
            mask = mask | (mask1 & mask2 & mask3)
    return mask


def domain_isovalue(field, isovalue):
    """Generate a mask for an isovalue real-space domain.

    Args:
        field (ndarray): Real-space field data.
        isovalue (float): Isovalue for the truncation.

    Returns:
        ndarray: Boolean mask.
    """
    return np.abs(field) > isovalue


def domain_sphere(obj, radius, centers=None):
    """Generate a mask for a spherical real-space domain.

    Args:
        obj: Atoms or SCF object.
        radius (float): Radius of the sphere.

    Keyword Args:
        centers (ndarray): Center of the sphere.

            Defaults to the geometric center of mass of the system. For multiple coordinates,
            multiple domains will be merged.

    Returns:
        ndarray: Boolean mask.
    """
    atoms = obj._atoms

    if centers is None:
        centers = center_of_mass(atoms.X)
    centers = np.asarray(centers)
    if centers.ndim == 1:
        mask = norm(centers - atoms.r, axis=1) < radius
    else:
        mask = np.zeros(atoms.Ns, dtype=bool)
        for center in centers:
            mask_tmp = norm(center - atoms.r, axis=1) < radius
            mask = mask | mask_tmp
    return mask


def truncate(field, mask):
    """Truncate field data for a given mask.

    This will not return a smaller array but set all truncated values to zero.

    Args:
        field (ndarray): Real-space field data.
        mask (ndarray): Boolean mask.

    Returns:
        ndarray: Boolean mask.
    """
    field_trunc = np.copy(field)
    field_trunc[~mask] = 0
    return field_trunc



#!/usr/bin/env python3
"""Implementation of different exchange-correlation functionals."""
from .utils import (
    ALIAS,
    get_exc,
    get_vxc,
    get_xc,
    get_zeta,
    IMPLEMENTED,
    parse_functionals,
    parse_xc_type,
    XC_MAP,
)

__all__ = ['ALIAS', 'get_exc', 'get_vxc', 'get_xc', 'get_zeta', 'IMPLEMENTED', 'parse_functionals',
           'parse_xc_type', 'XC_MAP']



#!/usr/bin/env python3
"""Various tools to check physical properties."""
import numpy as np
from numpy.linalg import norm
from scipy.optimize import minimize_scalar

from .dft import get_epsilon
from .gga import get_grad_field, get_tau
from .logger import log


def cutoff2gridspacing(E):
    """Convert plane wave energy cut-off to a real-space grid spacing.

    Reference: Phys. Rev. B 54, 14362.

    Args:
        E (float): Energy in Hartree.

    Returns:
        float: Grid spacing in Bohr.
    """
    return np.pi / np.sqrt(2 * E)


def gridspacing2cutoff(h):
    """Convert real-space grid spacing to plane wave energy cut-off.

    Reference: Phys. Rev. B 54, 14362.

    Args:
        h (float): Grid spacing in Bohr.

    Returns:
        float: Cut-off in Hartree.
    """
    return 0.5 * (np.pi / h)**2


def center_of_mass(coords, masses=None):
    """Calculate the center of mass for a set of coordinates and masses.

    Args:
        coords (ndarray): Array of real-space coordinates.

    Keyword Args:
        masses (ndarray): Mass or weight for each coordinate.

    Returns:
        ndarray: Center of mass.
    """
    if masses is None:
        masses = np.ones(len(coords))

    return np.sum(masses * coords.T, axis=1) / np.sum(masses)


def orbital_center(obj, psirs):
    """Calculate the orbital center of masses, e.g., from localized orbitals.

    Args:
        obj: Atoms or SCF object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        bool: Center of masses.
    """
    atoms = obj._atoms

    coms = [np.array([])] * 2
    Ncom = psirs.shape[2]
    for spin in range(atoms.occ.Nspin):
        coms_spin = np.empty((Ncom, 3))

        # Squared orbitals
        psi2 = np.real(psirs[spin].conj() * psirs[spin])
        for i in range(Ncom):
            coms_spin[i] = center_of_mass(atoms.r, psi2[:, i])
        coms[spin] = coms_spin
    return coms


def inertia_tensor(coords, masses=None):
    """Calculate the inertia tensor for a set of coordinates and masses.

    Reference: https://en.wikipedia.org/wiki/Moment_of_inertia

    Args:
        coords (ndarray): Array of real-space coordinates.

    Keyword Args:
        masses (ndarray): Mass or weight for each coordinate.

    Returns:
        ndarray: Inertia tensor.
    """
    if masses is None:
        masses = np.ones(len(coords))

    # The inertia tensor for a set of point masses can be calculated with a simple summation
    # https://en.wikipedia.org/wiki/Moment_of_inertia#Definition_2
    I = np.empty((3, 3))
    I[0, 0] = np.sum(masses * (coords[:, 1]**2 + coords[:, 2]**2))
    I[1, 1] = np.sum(masses * (coords[:, 0]**2 + coords[:, 2]**2))
    I[2, 2] = np.sum(masses * (coords[:, 0]**2 + coords[:, 1]**2))
    I[0, 1] = I[1, 0] = -np.sum(masses * (coords[:, 0] * coords[:, 1]))
    I[0, 2] = I[2, 0] = -np.sum(masses * (coords[:, 0] * coords[:, 2]))
    I[1, 2] = I[2, 1] = -np.sum(masses * (coords[:, 1] * coords[:, 2]))
    return I


def get_dipole(scf, n=None):
    """Calculate the electric dipole moment.

    This function does not account for periodicity, it may be a good idea to center the system.

    Reference: J. Chem. Phys. 155, 224109.

    Args:
        scf: SCF object.

    Keyword Args:
        n (float): Real-space electronic density.

    Returns:
        ndarray: Electric dipole moment in e * Bohr.
    """
    atoms = scf.atoms
    if n is None:
        if not hasattr(scf, 'n'):
            log.error('There is no density to calculate a dipole moment.')
            return 0
        n = scf.n

    # Diple moment: mu = \sum Z X - \int n(r) r dr
    mu = np.array([0, 0, 0], dtype=float)
    for i in range(atoms.Natoms):
        mu += atoms.Z[i] * atoms.X[i]

    for dim in range(3):
        mu[dim] -= atoms.dV * np.sum(n * atoms.r[:, dim])
    return mu


def get_ip(scf):
    """Calculate the ionization potential by calculating the negative HOMO energy.

    Reference: Physica 1, 104.

    Args:
        scf: SCF object.

    Returns:
        float: Ionization potential in Hartree.
    """
    epsilon = get_epsilon(scf, scf.W)
    # Add-up spin states
    epsilon = np.sum(epsilon, axis=0)
    return -epsilon[-1]


def check_ortho(obj, func, eps=1e-9):
    """Check the orthogonality condition for a set of functions.

    Args:
        obj: Atoms or SCF object.
        func (ndarray): A discretized set of functions.

    Keyword Args:
        eps (float): Tolerance for the condition.

    Returns:
        bool: Orthogonality status for the set of functions.
    """
    atoms = obj._atoms
    func = np.atleast_3d(func)

    # It makes no sense to calculate anything for only one function
    if atoms.occ.Nstate == 1:
        log.warning('Need at least two functions to check their orthogonality.')
        return True

    ortho_bool = True
    # Check the condition for every combination
    # Orthogonality condition: \int func1^* func2 dr = 0
    for spin in range(atoms.occ.Nspin):
        for i in range(atoms.occ.Nstate):
            for j in range(i + 1, atoms.occ.Nstate):
                res = atoms.dV * np.sum(func[spin, :, i].conj() * func[spin, :, j])
                tmp_bool = abs(res) < eps
                ortho_bool *= tmp_bool
                log.debug(f'Function {i} and {j}:\nValue: {res:.7f}\nOrthogonal: {tmp_bool}')
    log.info(f'Orthogonal: {ortho_bool}')
    return ortho_bool


def check_norm(obj, func, eps=1e-9):
    """Check the normalization condition for a set of functions.

    Args:
        obj: Atoms or SCF object.
        func (ndarray): A discretized set of functions.

    Keyword Args:
        eps (float): Tolerance for the condition.

    Returns:
        bool: Normalization status for the set of functions.
    """
    atoms = obj._atoms
    func = np.atleast_3d(func)

    norm_bool = True
    # Check the condition for every function
    # Normality condition: \int func^* func dr = 1
    for spin in range(atoms.occ.Nspin):
        for i in range(atoms.occ.Nstate):
            res = atoms.dV * np.sum(func[spin, :, i].conj() * func[spin, :, i])
            tmp_bool = abs(1 - res) < eps
            norm_bool *= tmp_bool
            log.debug(f'Function {i}:\nValue: {res:.7f}\nNormalized: {tmp_bool}')
    log.info(f'Normalized: {norm_bool}')
    return norm_bool


def check_orthonorm(obj, func):
    """Check the orthonormality conditions for a set of functions.

    Args:
        obj: Atoms or SCF object.
        func (ndarray): A discretized set of functions.

    Returns:
        bool: Orthonormality status for the set of functions.
    """
    atoms = obj._atoms
    ortho_bool = check_ortho(atoms, func)
    norm_bool = check_norm(atoms, func)
    log.info(f'Orthonormal: {ortho_bool * norm_bool}')
    return ortho_bool * norm_bool


def get_isovalue(n, percent=85):
    """Find an isovalue that contains a percentage of the electronic density.

    Reference: J. Chem. Phys. 158, 164102.

    Args:
        n (float): Real-space electronic density.

    Keyword Args:
        percent (float): Amount of density that should be contained.

    Returns:
        float: Isovalue that contains the specified percentage of the density.
    """
    def deviation(isovalue):
        """Wrapper function for finding the isovalue by minimization."""
        n_mask = np.sum(n[n > isovalue])
        return abs(percent - (n_mask / n_ref) * 100)

    # Integrated density
    n_ref = np.sum(n)
    # Finding the isovalue is an optimization problem, minimizing the deviation above
    # The problem is bound by zero (no density) and the maximum value in n
    res = minimize_scalar(deviation, bounds=(0, np.max(n)), method='bounded')
    return res.x


def get_tautf(scf):
    """Calculate the Thomas-Fermi kinetic energy densities per spin.

    Reference: Phys. Lett. B 63, 395.

    Args:
        scf: SCF object.

    Returns:
        ndarray: Real-space Thomas-Fermi kinetic energy density.
    """
    atoms = scf.atoms
    # Use the definition with a division by two
    tautf = 3 / 10 * (atoms.occ.Nspin * 3 * np.pi**2)**(2 / 3) * scf.n_spin**(5 / 3)

    log.debug(f'Calculated Ekin:  {scf.energies.Ekin:.6f} Eh')
    log.debug(f'Integrated tautf: {np.sum(tautf) * atoms.dV:.6f} Eh')
    return tautf


def get_tauw(scf):
    """Calculate the von Weizsaecker kinetic energy densities per spin.

    Reference: Z. Phys. 96, 431.

    Args:
        scf: SCF object.

    Returns:
        ndarray: Real-space von Weizsaecker kinetic energy density.
    """
    atoms = scf.atoms
    if scf.dn_spin is None:
        dn_spin = get_grad_field(atoms, scf.n_spin)
    else:
        dn_spin = scf.dn_spin
    dn2 = norm(dn_spin, axis=2)**2
    # Use the definition with a division by two
    tauw = dn2 / (8 * scf.n_spin)

    # For one- and two-electron systems the integrated KED has to be the same as the calculated KE
    log.debug(f'Calculated Ekin: {scf.energies.Ekin:.6f} Eh')
    log.debug(f'Integrated tauw: {np.sum(tauw) * atoms.dV:.6f} Eh')
    return tauw


def get_elf(scf):
    """Calculate the electron localization function.

    Reference: J. Chem. Phys. 92, 5397.

    Args:
        scf: SCF object.

    Returns:
        ndarray: Real-space electron localization function.
    """
    D = get_tau(scf.atoms, scf.Y) - get_tauw(scf)
    D0 = get_tautf(scf)
    X = D / D0
    return 1 / (1 + X**2)


def get_reduced_gradient(scf, eps=0):
    """Calculate the reduced density gradient s.

    Reference: Phys. Rev. Lett. 78, 1396.

    Args:
        scf: SCF object.

    Kwargs:
        eps (float): Threshold of the density where s should be truncated.

    Returns:
        ndarray: Real-space reduced density gradient.
    """
    atoms = scf.atoms
    if scf.dn_spin is None:
        dn_spin = get_grad_field(atoms, scf.n_spin)
    else:
        dn_spin = scf.dn_spin
    norm_dn = norm(np.sum(dn_spin, axis=0), axis=1)

    kf = (3 * np.pi**2 * scf.n)**(1 / 3)
    with np.errstate(divide='ignore', invalid='ignore'):
        s = norm_dn / (2 * kf * scf.n)
    s[scf.n < eps] = 0
    return s


def get_spin_squared(scf):
    """Calculate the expectation value of the squared spin operator <S^2>.

    Reference: Appl. Phys. Express 12, 115506.

    Args:
        scf: SCF object.

    Returns:
        float: The DFT value for <S^2>.
    """
    atoms = scf.atoms
    # <S^2> for a restricted calculation is always zero
    if not atoms.unrestricted:
        return 0

    rhoXr = scf.n_spin[0] - scf.n_spin[1]
    rhoXr[rhoXr < 0] = 0
    rhoX = np.sum(rhoXr) * atoms.dV
    SX = 0.5 * (np.sum(scf.n_spin[0]) - np.sum(scf.n_spin[1])) * atoms.dV
    return SX * (SX + 1) + rhoX


def get_multiplicity(scf):
    """Calculate the multiplicity from <S^2>.

    Args:
        scf: SCF object.

    Returns:
        float: Multiplicity 2S+1.
    """
    S2 = get_spin_squared(scf)
    # <S^2> = S(S+1) = S^2+S+0.25-0.25 = (S+0.5)^2-0.25 => S = sqrt(<S^2>+0.25)-0.5
    S = np.sqrt(S2 + 0.25) - 0.5
    return 2 * S + 1



#!/usr/bin/env python3
"""CUBE file handling."""
import textwrap
import time

import numpy as np

from ..data import NUMBER2SYMBOL, SYMBOL2NUMBER
from ..logger import log
from ..version import __version__


def read_cube(filename):
    """Load atom and cell data from CUBE files.

    There is no standard for CUBE files. The following format has been used.
    File format definition: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html

    Args:
        filename (str): CUBE input file path/name.

    Returns:
        tuple[list, ndarray, float, ndarray, int, ndarray]:
        Species, positions, charges, cell size, sampling, and field array.
    """
    if not filename.endswith('.cube'):
        filename += '.cube'

    # Atomic units and a cuboidal cell that starts at (0,0,0) are assumed.
    with open(filename, 'r') as fh:
        lines = fh.readlines()

        # The first and second lines can contain comments, print them if available
        comment = f'{lines[0].strip()}\n{lines[1].strip()}'
        if comment:
            log.info(f'CUBE file comment: "{comment}"')

        # Lines 4 to 6 contain the sampling per axis and the cell basis vectors with length a/s
        # A cuboidal cell is assumed, so only use the diagonal entries
        s = np.empty(3, dtype=int)
        a = np.empty(3)
        for i, line in enumerate(lines[3:6]):
            line_split = line.strip().split()
            s[i] = line_split[0]
            a[i] = s[i] * np.float_(line_split[i + 1])

        atom = []
        X = []
        Z = []
        # Following lines contain atom positions with the format: atom-id charge x-pos y-pos z-pos
        for _offset, line in enumerate(lines[6:]):
            line_split = line.strip().split()
            # If the first value is not a (positive) integer, we have reached the field data
            if not line_split[0].isdigit():
                break
            atom.append(NUMBER2SYMBOL[int(line_split[0])])
            Z.append(float(line_split[1]))
            X.append(np.float_(line_split[2:5]))
    X = np.asarray(X)

    # The rest of the data is the field data
    # Split the strings, flatten the lists of lists, and convert to a float numpy array
    tmp_list = [l.split() for l in lines[6 + _offset:]]
    field_list = [item for sublist in tmp_list for item in sublist]
    field = np.asarray(field_list, dtype=float)
    return atom, X, Z, a, s, field


def write_cube(obj, filename, field, fods=None, elec_symbols=('X', 'He')):
    """Generate CUBE files from given field data.

    There is no standard for CUBE files. The following format has been used to work with VESTA.
    File format definition: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html

    Args:
        obj: Atoms or SCF object.
        filename (str): CUBE output file path/name.
        field (ndarray): Real-space field data.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        None.
    """
    # Atomic units are assumed, so there is no need for conversion.
    atoms = obj._atoms

    if not filename.endswith('.cube'):
        filename += '.cube'

    if 'He' in atoms.atom and atoms.unrestricted:
        log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                    'polarized case.')

    # Make sure we have real-valued data in the correct order
    field = np.real(field)

    with open(filename, 'w') as fp:
        # The first two lines have to be a comment
        # Print information about the file and program, and the file creation time
        fp.write(f'File generated with eminus {__version__} on {time.ctime()}\n\n')
        # Number of atoms (int), and origin of the coordinate system (float)
        # The origin is normally at 0,0,0 but we could move our box, so take the minimum
        if fods is None:
            fp.write(f'{atoms.Natoms}  ')
        else:
            fp.write(f'{atoms.Natoms + sum([len(i) for i in fods])}  ')
        fp.write('0.0  0.0  0.0\n')
        # Number of points per axis (int), and vector defining the axis (float)
        # We only have a cuboidal box, so each vector only has one non-zero component
        fp.write(f'{atoms.s[0]}  {atoms.a[0] / atoms.s[0]:.6f}  0.0  0.0\n'
                 f'{atoms.s[1]}  0.0  {atoms.a[1] / atoms.s[1]:.6f}  0.0\n'
                 f'{atoms.s[2]}  0.0  0.0  {atoms.a[2] / atoms.s[2]:.6f}\n')
        # Atomic number (int), atomic charge (float), and atom position (floats) for every atom
        for ia in range(atoms.Natoms):
            fp.write(f'{SYMBOL2NUMBER[atoms.atom[ia]]}  {atoms.Z[ia]:.3f}  '
                     f'{atoms.X[ia, 0]: .6f}  {atoms.X[ia, 1]: .6f}  {atoms.X[ia, 2]: .6f}\n')
        if fods is not None:
            for s in range(len(fods)):
                for ie in fods[s]:
                    fp.write(f'{SYMBOL2NUMBER[elec_symbols[s]]}  0.000  '
                             f'{ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
        # Field data (float) with scientific formatting
        # We have s[0]*s[1] chunks values with a length of s[2]
        for i in range(atoms.s[0] * atoms.s[1]):
            # Print every round of values, so we can add empty lines between them
            data_str = '%+1.6e  ' * atoms.s[2] % tuple(field[i * atoms.s[2]:(i + 1) * atoms.s[2]])
            # Print a maximum of 6 values per row
            # Max width for this formatting is 90, since 6*len('+1.00000e-000  ')=90
            fp.write(f'{textwrap.fill(data_str, width=90)}\n\n')



#!/usr/bin/env python3
"""Perdew-Burke-Ernzerhof GGA exchange.

Reference: Phys. Rev. Lett. 78, 1396.
"""
import numpy as np
from numpy.linalg import norm

from .lda_x import lda_x, lda_x_spin


def gga_x_pbe(n, mu=0.2195149727645171, dn_spin=None, **kwargs):
    """Perdew-Burke-Ernzerhof parametrization of the exchange functional (spin-paired).

    Corresponds to the functional with the label GGA_X_PBE and ID 101 in Libxc.

    Reference: Phys. Rev. Lett. 78, 1396.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        mu (float): Functional parameter.
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE exchange energy density, potential, and vsigma.
    """
    ex, vx, _ = lda_x(n, **kwargs)
    gex, gvx, vsigmax = pbe_x_base(n, mu, dn_spin[0], **kwargs)
    return ex + gex / n, np.array([vx + gvx]), np.array([0.5 * vsigmax])


def gga_x_pbe_spin(n, zeta, mu=0.2195149727645171, dn_spin=None, **kwargs):
    """Perdew-Burke-Ernzerhof parametrization of the exchange functional (spin-polarized).

    Corresponds to the functional with the label GGA_X_PBE and ID 101 in Libxc.

    Reference: Phys. Rev. Lett. 78, 1396.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        mu (float): Functional parameter.
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE exchange energy density, potential, and vsigma.
    """
    # Use the spin-scaling relationship Exc(n_up, n_down)=(Exc(2 n_up)+Exc(2 n_down))/2
    zeta = zeta[0]  # Getting the non-zero values from zeta adds an extra dimension, remove it here
    n_up = zeta * n + n   # 2 * n_up
    n_dw = -zeta * n + n  # 2 * n_down
    ex_up, vx_up, vsigma_up = pbe_x_base(n_up, mu, 2 * dn_spin[0], **kwargs)
    ex_dw, vx_dw, vsigma_dw = pbe_x_base(n_dw, mu, 2 * dn_spin[1], **kwargs)

    ex, vx, _ = lda_x_spin(n, zeta, **kwargs)

    vsigmax = np.array([vsigma_up, np.zeros_like(ex), vsigma_dw])
    return ex + 0.5 * (ex_up + ex_dw) / n, np.array([vx[0] + vx_up, vx[1] + vx_dw]), vsigmax


def pbe_x_base(n, mu=0.2195149727645171, dn=None, **kwargs):
    """Base PBE exchange functional to be used in the spin-paired and -polarized case.

    Reference: Phys. Rev. Lett. 78, 1396.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        mu (float): Functional parameter.
        dn (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE exchange energy density, potential, and vsigma.
    """
    kappa = 0.804

    norm_dn = norm(dn, axis=1)
    kf = (3 * np.pi**2 * n)**(1 / 3)
    # Handle divisions by zero
    # divkf = 1 / kf
    divkf = np.divide(1, kf,
                      out=np.zeros_like(kf), where=(kf > 0))
    # Handle divisions by zero
    # s = norm_dn * divkf / (2 * n)
    s = np.divide(norm_dn * divkf, 2 * n,
                  out=np.zeros_like(n), where=(n > 0))
    f1 = 1 + mu * s**2 / kappa
    Fx = kappa - kappa / f1
    exunif = -3 * kf / (4 * np.pi)
    # In Fx a '1 + ' is missing, since n * exunif is the Slater exchange that is added later
    sx = exunif * Fx

    dsdn = -4 / 3 * s
    dFxds = 2 * mu * s / f1**2
    dexunif = exunif / 3
    exunifdFx = exunif * dFxds
    vx = sx + dexunif * Fx + exunifdFx * dsdn  # dFx/dn = dFx/ds * ds/dn

    # Handle divisions by zero
    # vsigmax = exunifdFx * divkf / (2 * norm_dn)
    vsigmax = np.divide(exunifdFx * divkf, 2 * norm_dn,
                        out=np.zeros_like(norm_dn), where=(norm_dn > 0))
    return sx * n, np.array([vx]), vsigmax



#!/usr/bin/env python3
"""Chachiyo GGA exchange.

Reference: Molecules 25, 3485.
"""
import numpy as np
from numpy.linalg import norm

from .lda_x import lda_x


def gga_x_chachiyo(n, dn_spin=None, **kwargs):
    """Chachiyo parametrization of the exchange functional (spin-paired).

    Corresponds to the functional with the label GGA_X_CHACHIYO and ID 298 in Libxc.

    Reference: Molecules 25, 3485.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: Chachiyo exchange energy density, potential, and vsigma.
    """
    norm_dn = norm(dn_spin[0], axis=1)
    ex, _, _ = lda_x(n, **kwargs)

    x = norm_dn / n**(4 / 3) * 2 / 9 * (np.pi / 3)**(1 / 3)
    x1 = x + 1
    logx1 = np.log(x1)
    div = 3 * x + np.pi**2
    tmpgex = 3 * x**2 + np.pi**2 * logx1
    gex = tmpgex / (div * logx1)

    term1 = 8 * ex / tmpgex * (x**2 + x * np.pi**2 / (6 * x1)) + \
        2 / 3 * norm_dn / n * (1 / div + 1 / (3 * logx1 * x1))
    gvx = (1 + 1 / 3) * ex - term1

    vsigmax = n * 3 * term1 / (8 * norm_dn**2)
    return ex * gex, np.array([gvx]) * gex, np.array([vsigmax]) * gex


def gga_x_chachiyo_spin(n, zeta, dn_spin=None, **kwargs):
    """Chachiyo parametrization of the exchange functional (spin-polarized).

    Corresponds to the functional with the label GGA_X_CHACHIYO and ID 298 in Libxc.

    Reference: Molecules 25, 3485.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: Chachiyo exchange energy density, potential, and vsigma.
    """
    # Use the spin-scaling relationship Exc(n_up, n_down)=(Exc(2 n_up)+Exc(2 n_down))/2
    n_up = zeta * n + n   # 2 * n_up
    n_dw = -zeta * n + n  # 2 * n_down
    ex_up, vx_up, vsigma_up = gga_x_chachiyo(n_up, np.array([2 * dn_spin[0]]), **kwargs)
    ex_dw, vx_dw, vsigma_dw = gga_x_chachiyo(n_dw, np.array([2 * dn_spin[1]]), **kwargs)

    vsigmax = np.array([2 * vsigma_up, np.zeros_like(vsigma_up), 2 * vsigma_dw])
    return 0.5 * (ex_up * n_up + ex_dw * n_dw) / n, np.array([vx_up, vx_dw]), vsigmax



#!/usr/bin/env python3
"""XYZ file handling."""
import time

import numpy as np

from ..logger import log
from ..units import ang2bohr, bohr2ang
from ..version import __version__


def read_xyz(filename):
    """Load atom species and positions from XYZ files.

    File format definition: https://openbabel.org/wiki/XYZ_%28format%29

    Args:
        filename (str): XYZ input file path/name.

    Returns:
        tuple[list, ndarray]: Atom species and positions.
    """
    if not filename.endswith('.xyz'):
        filename += '.xyz'

    with open(filename, 'r') as fh:
        lines = fh.readlines()

        # The first line contains the number of atoms
        Natoms = int(lines[0].strip())

        # The second line can contain a comment, print it if available
        comment = lines[1].strip()
        if comment:
            log.info(f'XYZ file comment: "{comment}"')

        atom = []
        X = []
        # Following lines contain atom positions with the format: Atom x-pos y-pos z-pos
        for line in lines[2:2 + Natoms]:
            line_split = line.strip().split()
            atom.append(line_split[0])
            X.append(np.float_(line_split[1:4]))

    # xyz files are in Angstrom, so convert to Bohr
    X = ang2bohr(np.asarray(X))
    return atom, X


def write_xyz(obj, filename, fods=None, elec_symbols=('X', 'He'), trajectory=False):
    """Generate XYZ files from atoms objects.

    File format definition: https://openbabel.org/wiki/XYZ_%28format%29

    Args:
        obj: Atoms or SCF object.
        filename (str): XYZ output file path/name.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.
        trajectory (bool): Allow appending to a file to create trajectories.

    Returns:
        None.
    """
    atoms = obj._atoms

    if not filename.endswith('.xyz'):
        filename += '.xyz'

    # Convert the coordinates from atomic units to Angstrom
    X = bohr2ang(atoms.X)
    if fods is not None:
        fods = [bohr2ang(i) for i in fods]

    if 'He' in atoms.atom and atoms.unrestricted:
        log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                    'polarized case.')

    # Append to a file when using the trajectory keyword
    if trajectory:
        mode = 'a'
    else:
        mode = 'w'

    with open(filename, mode) as fp:
        # The first line contains the number of atoms
        # If we add FOD coordinates, add them to the count
        if fods is None:
            fp.write(f'{atoms.Natoms}\n')
        else:
            fp.write(f'{atoms.Natoms + sum([len(i) for i in fods])}\n')
        # The second line can contain a comment
        # Print information about the file and program, and the file creation time
        fp.write(f'File generated with eminus {__version__} on {time.ctime()}\n')
        for ia in range(atoms.Natoms):
            fp.write(f'{atoms.atom[ia]:<2s}  {X[ia, 0]: .6f}  {X[ia, 1]: .6f}  {X[ia, 2]: .6f}\n')
        # Add FOD coordinates if desired
        # The atom symbol will default to X (no atom type)
        if fods is not None:
            for s in range(len(fods)):
                for ie in fods[s]:
                    fp.write(f'{elec_symbols[s]:<2s}  {ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')



#!/usr/bin/env python3
"""Workflow functions that combine functions to generate orbitals."""
from .dft import get_psi
from .io import write_cube
from .localizer import get_FLO, get_FO, get_wannier
from .logger import log


def KSO(scf, write_cubes=False, **kwargs):
    """Generate Kohn-Sham orbitals and optionally save them as CUBE files.

    Reference: Phys. Rev. 140, A1133.

    Args:
        scf: SCF object.

    Keyword Args:
        write_cubes (bool): Write orbitals to CUBE files.
        **kwargs: Throwaway arguments.

    Returns:
        ndarray: Real-space Kohn-Sham orbitals.
    """
    atoms = scf.atoms

    # Calculate eigenfunctions and transform to real-space
    kso = atoms.I(get_psi(scf, scf.W))
    if write_cubes:
        cube_writer(atoms, 'KSO', kso)
    return kso


def FO(scf, write_cubes=False, fods=None):
    """Generate Fermi orbitals and optionally save them as CUBE files.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        scf: SCF object.

    Keyword Args:
        write_cubes (bool): Write orbitals to CUBE files.
        fods (list): Fermi-orbital descriptors.

    Returns:
        ndarray: Real-space Fermi orbitals.
    """
    # Lazy import extras
    from .extras.fods import get_fods, remove_core_fods
    atoms = scf.atoms

    # Calculate eigenfunctions
    kso = get_psi(scf, scf.W)
    if fods is None:
        fods = get_fods(atoms)
    fods = remove_core_fods(atoms, fods)

    # The FO functions need orbitals in reciprocal space as input
    fo = get_FO(atoms, kso, fods)
    if write_cubes:
        cube_writer(atoms, 'FO', fo)
    return fo


def FLO(scf, write_cubes=False, fods=None):
    """Generate Fermi-Loewdin orbitals and optionally save them as CUBE files.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        scf: SCF object.

    Keyword Args:
        write_cubes (bool): Write orbitals to CUBE files.
        fods (list): Fermi-orbital descriptors.

    Returns:
        ndarray: Real-space Fermi-Loewdin orbitals.
    """
    # Lazy import extras
    from .extras.fods import get_fods, remove_core_fods
    atoms = scf.atoms

    # Calculate eigenfunctions
    kso = get_psi(scf, scf.W)
    if fods is None:
        fods = get_fods(atoms)
    fods = remove_core_fods(atoms, fods)

    # The FLO functions need orbitals in reciprocal space as input
    flo = get_FLO(atoms, kso, fods)
    if write_cubes:
        cube_writer(atoms, 'FLO', flo)
    return flo


def WO(scf, write_cubes=False, precondition=True):
    """Generate Wannier orbitals and optionally save them as CUBE files.

    Reference: Phys. Rev. B 59, 9703.

    Args:
        scf: SCF object.

    Keyword Args:
        write_cubes (bool): Write orbitals to CUBE files.
        precondition (bool): Precondition by calculating FLOs as the initial guess.

    Returns:
        ndarray: Real-space Wannier orbitals.
    """
    atoms = scf.atoms

    # Calculate eigenfunctions/initial guess orbitals and transform to real-space
    if precondition:
        psi = FLO(scf)
    else:
        psi = atoms.I(get_psi(scf, scf.W))

    wo = get_wannier(atoms, psi)
    if write_cubes:
        cube_writer(atoms, 'WO', wo)
    return wo


def cube_writer(atoms, orb_type, orbitals):
    """Simple CUBE file writer function.

    Args:
        atoms: Atoms object.
        orb_type (str): Orbital orb_type for the CUBE file names.
        orbitals (ndarray): Real-space orbitals.
    """
    # Create the system name
    name = ''
    for ia in sorted(set(atoms.atom)):
        # Skip the number of atoms if it is equal to one
        na = atoms.atom.count(ia)
        name += f'{ia}{na if na > 1 else ""}'

    n_spin = ''
    for spin in range(atoms.occ.Nspin):
        for i in range(atoms.occ.Nstate):
            if atoms.occ.f[spin, i] > 0:
                if atoms.occ.Nspin > 1:
                    n_spin = f'_spin_{spin}'
                filename = f'{name}_{orb_type}_{i}{n_spin}.cube'
                log.info(f'Write {filename}...')
                write_cube(atoms, filename, orbitals[spin, :, i])



#!/usr/bin/env python3
"""SCF class definition."""
import copy
import logging
import time

import numpy as np

from .dft import get_epsilon, guess_pseudo, guess_random
from .energies import Energy, get_Edisp, get_Eewald, get_Esic
from .gth import GTH
from .logger import create_logger, get_level
from .minimizer import IMPLEMENTED as all_minimizer
from .potentials import IMPLEMENTED as all_potentials
from .potentials import init_pot
from .tools import center_of_mass, get_spin_squared
from .version import info
from .xc import parse_functionals, parse_xc_type


class SCF:
    """Perform direct minimizations.

    Args:
        atoms: Atoms object.

    Keyword Args:
        xc (str): Comma-separated exchange-correlation functional description.

            Adding 'libxc:' before a functional will use the Libxc interface for that functional,
            e.g., with :code:`libxc:mgga_x_scan,libxc:mgga_c_scan`.
        pot (str): Type of potential.

            Can be one of 'GTH' (default), 'Coulomb', 'Harmonic', or 'Ge'. Alternatively, a path to
            a directory containing custom GTH pseudopotential files can be given.
        guess (str): Initial guess for the wave functions.

            Can be one of 'random' (default) or 'pseudo'. Adding 'symm' to the string will use the
            same coefficients for both spin channels, e.g., :code:`symm-rand`.
        etol (float): Convergence tolerance of the total energy.
        gradtol (float): Convergence tolerance of the gradient norm.

            This tolerance will only be used in conjugate gradient methods.
        sic (bool): Calculate the Kohn-Sham Perdew-Zunger SIC energy at the end of the SCF step.
        disp (bool | dict): Calculate a dispersion correction.

            A dictionary can be used to pass arguments to the respective
            function, e.g., with :code:`{'version': 'd3zero', 'atm': False, 'xc': 'scan'}`.
        opt (dict | None): Dictionary to customize the minimization methods.

            The keys can be chosen out of 'sd', 'lm', 'pclm', 'cg', 'pccg', and 'auto'. Defaults to
            :code:`{'auto': 250}`.
        verbose (int | str): Level of output.

            Can be one of 'critical', 'error', 'warning', 'info' (default), or 'debug'. An integer
            value can be used as well, where larger numbers mean more output, starting from 0.
            Defaults to the verbosity level of the Atoms object.
    """
    def __init__(self, atoms, xc='lda,vwn', pot='gth', guess='random', etol=1e-7, gradtol=None,
                 sic=False, disp=False, opt=None, verbose=None):
        """Initialize the SCF object."""
        # Set opt here, better to not use mutable data types in signatures
        if opt is None:
            opt = {'auto': 250}

        # Set the input parameters (the ordering is important)
        self.atoms = atoms              #: Atoms object.
        self.log = create_logger(self)  #: Logger object.
        self.verbose = verbose          #: Verbosity level.
        self.xc = xc                    #: Exchange-correlation functional.
        self.pot = pot                  #: Used potential.
        self.guess = guess              #: Initial wave functions guess.
        self.etol = etol                #: Total energy convergence tolerance.
        self.gradtol = gradtol          #: Gradient norm convergence tolerance.
        self.sic = sic                  #: Enables the SIC energy calculation.
        self.disp = disp                #: Enables the dispersion correction calculation.
        self.opt = opt                  #: Minimization methods.

        # Initialize other attributes
        self.energies = Energy()        #: Energy object holding energy contributions.
        self.is_converged = False       #: Determines the SCF object convergence.

    # ### Class properties ###

    @property
    def atoms(self):
        """Atoms object."""
        return self._atoms

    @atoms.setter
    def atoms(self, value):
        # Build the Atoms object if necessary and make a copy
        # This way the Atoms objects inside and outside the class are independent but both are build
        if not value.is_built:
            self._atoms = copy.deepcopy(value.build())
        else:
            self._atoms = copy.deepcopy(value)

    @property
    def xc(self):
        """Exchange-correlation functional."""
        return self._xc

    @xc.setter
    def xc(self, value):
        self._xc = parse_functionals(value.lower())
        # Determine the type of functional combinations
        self._xc_type = parse_xc_type(self._xc)
        if 'mock_xc' in self._xc:
            self.log.warning('Usage of mock functional detected.')

    @property
    def pot(self):
        """Used potential."""
        return self._pot

    @pot.setter
    def pot(self, value):
        if value.lower() in all_potentials:
            self._pot = value.lower()
            # Only set the pseudopotential type for GTH pseudopotentials
            if self._pot == 'gth':
                if 'gga' in self._xc_type:
                    self._psp = 'pbe'
                else:
                    self._psp = 'pade'
        # If pot is no supported potential treat it as a path to a directory containing GTH files
        else:
            self.log.info(f'Use the path "{value}" to search for GTH pseudopotential files.')
            self._psp = value
            self._pot = 'gth'
        # Build the potential
        if self._pot == 'gth':
            self.gth = GTH(self)
        self.Vloc = init_pot(self)

    @property
    def guess(self):
        """Initial wave functions guess."""
        return self._guess

    @guess.setter
    def guess(self, value):
        # Set the guess method
        value = value.lower()
        if 'rand' in value:
            self._guess = 'random'
        elif 'pseudo' in value:
            self._guess = 'pseudo'
        else:
            ValueError(f'{value} is no valid initial guess.')
        # Check if a symmetric or unsymmetric guess is selected
        if 'sym' in value and 'unsym' not in value:
            self._symmetric = True
        else:
            self._symmetric = False

    @property
    def opt(self):
        """Minimization methods."""
        return self._opt

    @opt.setter
    def opt(self, value):
        # Set lowercase to all keys
        value = {k.lower(): v for k, v in value.items()}
        for opt in value:
            if opt not in all_minimizer:
                KeyError(f'No minimizer found for "{opt}".')
        self._opt = value

    @property
    def verbose(self):
        """Verbosity level."""
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        # If no verbosity is given use the one from the Atoms object
        if value is None:
            value = self.atoms.verbose
        self._verbose = get_level(value)
        self.log.verbose = self._verbose

    # ### Read-only properties ###

    @property
    def psp(self):
        """Pseudopotential path."""
        return self._psp

    @property
    def symmetric(self):
        """Determines if the initial guess is the same for both spin channels."""
        return self._symmetric

    @property
    def xc_type(self):
        """Determines the exchange-correlation family."""
        return self._xc_type

    # ### Class methods ###

    def run(self, **kwargs):
        """Run the self-consistent field (SCF) calculation."""
        # Print some information about the calculation
        if self.log.level <= logging.DEBUG:
            info()
        self.log.debug(f'\n--- Atoms information ---\n{self.atoms}\n'
                       f'\n--- Cell information ---\nSide lengths: {self.atoms.a} a0\n'
                       f'Sampling per axis: {self.atoms.s}\n'
                       f'Cut-off energy: {self.atoms.ecut} Eh\n'
                       f'Compression: {len(self.atoms.G2c) / len(self.atoms.G2):.5f}\n'
                       f'\n--- State information ---\n{self.atoms.occ}\n'
                       f'\n--- Calculation information ---\n{self}\n\n--- SCF data ---')

        # Calculate Ewald energy that only depends on the system geometry
        self.energies.Eewald = get_Eewald(self.atoms)
        # Build the initial wave function if there is no W to start from
        if not hasattr(self, 'W'):
            if 'random' in self.guess:
                self.W = guess_random(self, symmetric=self.symmetric)
            elif 'pseudo' in self.guess:
                self.W = guess_pseudo(self, symmetric=self.symmetric)

        # Start the minimization procedures
        self.clear()
        Etots = []
        for imin in self.opt:
            # Call the minimizer
            self.log.info(f'Start {all_minimizer[imin].__name__}...')
            start = time.perf_counter()
            Elist = all_minimizer[imin](self, self.opt[imin], **kwargs)
            end = time.perf_counter()
            # Save the minimizer results
            self._opt_log[imin] = {}
            self._opt_log[imin]['iter'] = len(Elist)
            self._opt_log[imin]['time'] = end - start
            Etots += Elist
            # Do not start other minimizations if one converged
            if self.is_converged:
                break
        if self.is_converged:
            self.log.info(f'SCF converged after {len(Etots)} iterations.')
        else:
            self.log.warning('SCF not converged!')

        # Calculate SIC energy if desired
        if self.sic:
            self.energies.Esic = get_Esic(self, self.Y)
        # Calculate dispersion correction energy if desired
        if isinstance(self.disp, dict):
            self.energies.Edisp = get_Edisp(self, **self.disp)
        elif self.disp:
            self.energies.Edisp = get_Edisp(self)

        # Print minimizer timings
        self.log.debug('\n--- SCF results ---')
        t_tot = 0
        for imin in self._opt_log:
            N = self._opt_log[imin]['iter']
            t = self._opt_log[imin]['time']
            t_tot += t
            self.log.debug(f'Minimizer: {imin}'
                           f'\nIterations: {N}'
                           f'\nTime: {t:.5f} s'
                           f'\nTime/Iteration: {t / N:.5f} s')
        self.log.info(f'Total SCF time: {t_tot:.5f} s')
        # Print the S^2 expectation value for unrestricted calculations
        if self.atoms.unrestricted:
            self.log.info(f'<S^2> = {get_spin_squared(self):.6e}')
        # Print energy data
        if self.log.level <= logging.DEBUG:
            self.log.debug('\n--- Energy data ---\n'
                           f'Eigenenergies:\n{get_epsilon(self, self.W)}\n\n{self.energies}')
        else:
            self.log.info(f'Etot = {self.energies.Etot:.9f} Eh')
        return self.energies.Etot

    kernel = run

    def recenter(self, center=None):
        """Recenter the system inside the cell.

        Keyword Args:
            center (float | list | tuple | ndarray | None): Point to center the system around.
        """
        atoms = self.atoms
        # Get the COM before centering the atoms
        com = center_of_mass(atoms.X)
        # Run the recenter method of the atoms object
        self.atoms.recenter(center=center)
        if center is None:
            dr = com - atoms.a / 2
        else:
            center = np.asarray(center)
            dr = com - center

        # Shift orbitals and density
        if hasattr(self, 'W'):
            self.W = atoms.T(self.W, dr=-dr)
        # Transform the density to the reciprocal space, shift, and transform back
        if hasattr(self, 'n'):
            Jn = atoms.J(self.n)
            TJn = atoms.T(Jn, dr=-dr)
            self.n = np.real(atoms.I(TJn))

        # Recalculate the potential since it depends on the structure factor
        self.pot = self.pot
        # Clear intermediate results to make sure no one uses the unshifted results
        self.clear()
        return self

    def clear(self):
        """Initialize or clear intermediate results."""
        self.Y = None           # Orthogonal wave functions
        self.n_spin = None      # Electronic densities per spin
        self.dn_spin = None     # Gradient of electronic densities per spin
        self.tau = None         # Kinetic energy densities per spin
        self.phi = None         # Hartree field
        self.exc = None         # Exchange-correlation energy density
        self.vxc = None         # Exchange-correlation potential
        self.vsigma = None      # n times d exc/d |dn|^2
        self.vtau = None        # d exc/d tau
        self._precomputed = {}  # Dictionary of pre-computed values not to be saved
        self._opt_log = {}      # Log of the optimization procedure
        return self

    def __repr__(self):
        """Print the most important parameters stored in the SCF object."""
        # Use chr(10) to create a linebreak since backslashes are not allowed in f-strings
        return f'XC functionals: {self.xc}\n' \
               f'Potential: {self.pot}\n' \
               f'{f"GTH files: {self.psp}" + chr(10) if self.pot == "gth" else ""}' \
               f'Starting guess: {self.guess}\n' \
               f'Symmetric guess: {self.symmetric}\n' \
               f'Energy convergence tolerance: {self.etol} Eh\n' \
               f'Gradient convergence tolerance: {self.gradtol}\n' \
               f'Non-local potential: {self.gth.NbetaNL > 0 if self.pot == "gth" else "false"}'


class RSCF(SCF):
    """SCF class for spin-paired systems.

    Inherited from :class:`eminus.scf.SCF`.

    In difference to the SCF class, this class will not build the original Atoms object, only the
    one attributed to the class. Customized fillings could be overwritten when using this class.
    """
    @property
    def atoms(self):
        """Atoms object."""
        return self._atoms

    @atoms.setter
    def atoms(self, value):
        self._atoms = copy.deepcopy(value)
        self._atoms.unrestricted = False
        self._atoms.occ.fill()
        if not self._atoms.is_built:
            self._atoms = self._atoms.build()


class USCF(SCF):
    """SCF class for spin-polarized systems.

    Inherited from :class:`eminus.scf.SCF`.

    In difference to the SCF class, this class will not build the original Atoms object, only the
    one attributed to the class. Customized fillings could be overwritten when using this class.
    """
    @property
    def atoms(self):
        """Atoms object."""
        return self._atoms

    @atoms.setter
    def atoms(self, value):
        self._atoms = copy.deepcopy(value)
        self._atoms.unrestricted = True
        self._atoms.occ.fill()
        if not self._atoms.is_built:
            self._atoms = self._atoms.build()



#!/usr/bin/env python3
"""Viewer functions for Jupyter notebooks."""
import io
import pathlib

import numpy as np

from ..data import COVALENT_RADII, CPK_COLORS
from ..io import create_pdb_str, read_cube, read_xyz
from ..logger import log
from ..tools import get_isovalue
from .fods import split_fods


def view(*args, **kwargs):
    """Unified display function."""
    if isinstance(args[0], str):
        return view_file(*args, **kwargs)
    return view_atoms(*args, **kwargs)


def view_atoms(obj, extra=None, plot_n=False, percent=85, surfaces=20):
    """Display atoms and 3D-coordinates, e.g., FODs or grid points, or even densities.

    Reference: https://plotly.com/python

    Args:
        obj: Atoms or SCF object.

    Keyword Args:
        extra (ndarray | list ): Extra coordinates or FODs to display.
        plot_n (bool): Weather to plot the electronic density (only for executed scf objects).
        percent (float): Amount of density that should be contained.
        surfaces (int): Number of surfaces to display in density plots (reduce for performance).

    Returns:
        None.
    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise
    atoms = obj._atoms

    fig = go.Figure()
    # Add species one by one to be able to have them named and be selectable in the legend
    # Note: The size scaling is mostly arbitrary and has no meaning
    for ia in sorted(set(atoms.atom)):
        mask = np.where(np.asarray(atoms.atom) == ia)[0]
        atom_data = go.Scatter3d(x=atoms.X[mask, 0], y=atoms.X[mask, 1], z=atoms.X[mask, 2],
                                 name=ia,
                                 mode='markers',
                                 marker={'size': 2 * np.pi * np.sqrt(COVALENT_RADII[ia]),
                                         'color': CPK_COLORS[ia],
                                         'line': {'color': 'black', 'width': 2}})
        fig.add_trace(atom_data)
    if extra is not None:
        # If a list has been provided with a length of two it has to be FODs
        if isinstance(extra, list):
            if len(extra[0]) != 0:
                extra_data = go.Scatter3d(x=extra[0][:, 0], y=extra[0][:, 1], z=extra[0][:, 2],
                                          name='up-FOD',
                                          mode='markers',
                                          marker={'size': np.pi, 'color': 'red'})
                fig.add_trace(extra_data)
            if len(extra) > 1 and len(extra[1]) != 0:
                extra_data = go.Scatter3d(x=extra[1][:, 0], y=extra[1][:, 1], z=extra[1][:, 2],
                                          name='down-FOD',
                                          mode='markers',
                                          marker={'size': np.pi, 'color': 'green'})
                fig.add_trace(extra_data)
        # Treat extra as normal coordinates otherwise
        else:
            extra_data = go.Scatter3d(x=extra[:, 0], y=extra[:, 1], z=extra[:, 2],
                                      name='Coordinates',
                                      mode='markers',
                                      marker={'size': 1, 'color': 'red'})
            fig.add_trace(extra_data)

    # A density can be plotted for an SCF object
    if isinstance(plot_n, np.ndarray) or plot_n:
        # Plot a given array or use the density from an SCF object
        if isinstance(plot_n, np.ndarray):
            density = plot_n
        else:
            density = obj.n
        den_data = go.Volume(x=atoms.r[:, 0], y=atoms.r[:, 1], z=atoms.r[:, 2], value=density,
                             name='Density',
                             colorbar_title=f'Density ({percent}%)',
                             colorscale='Inferno',
                             isomin=get_isovalue(density, percent=percent),
                             isomax=np.max(density),
                             surface_count=surfaces,
                             opacity=0.1,
                             showlegend=True)
        fig.add_trace(den_data)
        # Move colorbar to the left
        fig.data[-1].colorbar.x = -0.15

    # Theming
    fig.update_layout(scene={'xaxis': {'range': [0, atoms.a[0]], 'title': 'x [a<sub>0</sub>]'},
                             'yaxis': {'range': [0, atoms.a[1]], 'title': 'y [a<sub>0</sub>]'},
                             'zaxis': {'range': [0, atoms.a[2]], 'title': 'z [a<sub>0</sub>]'},
                             'aspectmode': 'cube'},
                      legend={'itemsizing': 'constant', 'title': 'Selection'},
                      hoverlabel_bgcolor='black',
                      template='none')
    fig.show()


def view_file(filename, isovalue=0.01, gui=False, elec_symbols=('X', 'He'), **kwargs):
    """Display molecules and orbitals.

    Reference: Bioinformatics 34, 1241.

    Args:
        filename (str): Input file path/name. This can be either a cube or xyz file.

    Keyword Args:
        isovalue (float): Isovalue for sizing orbitals.
        gui (bool): Turn on the NGLView GUI.
        elec_symbols (list): Identifier for up and down FODs.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        NGLWidget: Viewable object.
    """
    try:
        from nglview import NGLWidget, TextStructure, write_html
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise

    if isinstance(isovalue, str):
        isovalue = float(isovalue)

    view = NGLWidget(**kwargs)
    view._set_size('400px', '400px')
    # Set the gui to the view
    if gui:
        view.gui_style = 'ngl'

    # Handle XYZ files
    if filename.endswith('.xyz'):
        # Atoms
        atom, X = read_xyz(filename)
        atom, X, X_fod = split_fods(atom, X, elec_symbols)
        view.add_component(TextStructure(create_pdb_str(atom, X)))
        view[0].clear()
        view[0].add_ball_and_stick()
        # Spin up FODs
        if len(X_fod[0]) > 0:
            view.add_component(TextStructure(create_pdb_str([elec_symbols[0]] * len(X_fod[0]),
                               X_fod[0])))
            view[1].clear()
            view[1].add_ball_and_stick(f'_{elec_symbols[0]}', color='red', radius=0.1)
        # Spin down FODs
        if len(X_fod[1]) > 0:
            view.add_component(TextStructure(create_pdb_str([elec_symbols[1]] * len(X_fod[1]),
                               X_fod[1])))
            view[2].clear()
            view[2].add_ball_and_stick(f'_{elec_symbols[1]}', color='green', radius=0.1)
        view.center()
    # Handle CUBE files
    elif filename.endswith('.cube'):
        # Atoms and cell
        atom, X, _, a, _, _ = read_cube(filename)
        atom, X, X_fod = split_fods(atom, X, elec_symbols)
        view.add_component(TextStructure(create_pdb_str(atom, X, a)))
        view[0].clear()
        view[0].add_ball_and_stick()
        view.add_unitcell()
        view.center()
        # Spin up
        view.add_component(filename)
        view[1].clear()
        # Negate isovalue here as a workaround for some display bugs
        view[1].add_surface(negateIsolevel=True,
                            isolevelType='value',
                            isolevel=-isovalue,
                            color='lightgreen',
                            opacity=0.75,
                            side='front')
        # Spin down
        view.add_component(filename)
        view[2].clear()
        view[2].add_surface(negateIsolevel=True,
                            isolevelType='value',
                            isolevel=isovalue,
                            color='red',
                            opacity=0.75,
                            side='front')
        # Spin up FODs
        if len(X_fod[0]) > 0:
            view.add_component(TextStructure(create_pdb_str([elec_symbols[0]] * len(X_fod[0]),
                               X_fod[0])))
            view[3].clear()
            view[3].add_ball_and_stick(f'_{elec_symbols[0]}', color='red', radius=0.1)
        # Spin down FODs
        if len(X_fod[1]) > 0:
            view.add_component(TextStructure(create_pdb_str([elec_symbols[1]] * len(X_fod[1]),
                               X_fod[1])))
            view[4].clear()
            view[4].add_ball_and_stick(f'_{elec_symbols[1]}', color='green', radius=0.1)
    # Handle other files (mainly PDB)
    else:
        # If no xyz or cube file is used try a more generic file viewer
        ext = pathlib.Path(filename).suffix.replace('.', '')
        # It seems that only pdb works with this method
        if ext != 'pdb':
            log.warning('Only XYZ, CUBE, and PDB files are support, but others might work.')
        with open(filename, 'r') as fh:
            view.add_component(fh, ext=ext)
        view[0].clear()
        view[0].add_ball_and_stick()
        view.center()

    # Check if the function has been executed in a notebook
    # If yes, just return the view
    if executed_in_notebook():
        return view
    # Otherwise the viewer would display nothing
    # But if plotly is installed on can write the view to a StringIO object and display it with the
    # smart open_html_in_browser function from plotly that automatically opens a new browser tab
    try:
        from plotly.io._base_renderers import open_html_in_browser
        # Use StringIO object instead of a temporary file
        with io.StringIO() as html:
            write_html(html, view)
            open_html_in_browser(html.getvalue())
    except ImportError:
        log.error('This function only works in notebooks or with Plotly installed.')
    return None


def executed_in_notebook():
    """Check if the code runs in a notebook.

    Reference: https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook

    Returns:
        bool: Weather in a notebook or not.
    """
    try:
        shell = get_ipython().__class__.__name__  # type: ignore[name-defined]
        # True for Jupyter notebook or qtconsole
        # False for terminal running IPython or other types
        return shell == 'ZMQInteractiveShell'
    except NameError:
        return False



#!/usr/bin/env python3
"""Collection of miscellaneous potentials."""
import numpy as np
from scipy.linalg import norm

from .gth import init_gth_loc
from .logger import log


def harmonic(scf):
    """Harmonic potential.

    Can be used for quantum dot calculations.

    Args:
        scf: SCF object.

    Returns:
        ndarray: Harmonic potential in reciprocal space.
    """
    atoms = scf.atoms
    freq = 2
    dr = norm(atoms.r - np.sum(atoms.R, axis=1) / 2, axis=1)
    Vharm = 0.5 * freq**2 * dr**2
    return atoms.Jdag(atoms.O(atoms.J(Vharm)))


def coulomb(scf):
    """All-electron Coulomb potential.

    Reference: Bull. Lebedev Phys. Inst. 42, 329.

    Args:
        scf: SCF object.

    Returns:
        ndarray: Coulomb potential in reciprocal space.
    """
    atoms = scf.atoms
    Z = atoms.Z[0]  # This potential should only be used for same species

    # Ignore the division by zero for the first elements
    # One could do some proper indexing with [1:] but indexing is slow
    with np.errstate(divide='ignore', invalid='ignore'):
        Vcoul = -4 * np.pi * Z / atoms.G2
    Vcoul[0] = 0

    Sf = np.sum(atoms.Sf, axis=0)
    return atoms.J(Vcoul * Sf)


def ge(scf):
    """Starkloff-Joannopoulos local pseudopotential for germanium.

    Fourier-transformed by Tomas Arias.

    Reference: Phys. Rev. B 16, 5212.

    Args:
        scf: SCF object.

    Returns:
        ndarray: Germanium pseudopotential in reciprocal space.
    """
    atoms = scf.atoms
    Z = 4  # This potential should only be used for germanium
    lamda = 18.5
    rc = 1.052
    Gm = np.sqrt(atoms.G2)
    Vps = np.empty_like(atoms.G2)

    with np.errstate(divide='ignore', invalid='ignore'):
        Vps = -2 * np.pi * np.exp(-np.pi * Gm / lamda) * np.cos(rc * Gm) * (Gm / lamda) / \
            (1 - np.exp(-2 * np.pi * Gm / lamda))
        for n in range(5):
            Vps = Vps + (-1)**n * np.exp(-lamda * rc * n) / (1 + (n * lamda / Gm)**2)
        Vps = Vps * 4 * np.pi * Z / Gm**2 * (1 + np.exp(-lamda * rc)) - 4 * np.pi * Z / Gm**2

    # Special case for G=(0,0,0)
    n = np.arange(1, 5)
    Vps[0] = 4 * np.pi * Z * (1 + np.exp(-lamda * rc)) * \
        (rc**2 / 2 + 1 / lamda**2 * (np.pi**2 / 6 +
         np.sum((-1)**n * np.exp(-lamda * rc * n) / n**2)))

    Sf = np.sum(atoms.Sf, axis=0)
    return atoms.J(Vps * Sf)


def init_pot(scf):
    """Handle and initialize potentials.

    Args:
        scf: SCF object.

    Returns:
        ndarray: Potential in reciprocal space.
    """
    try:
        pot = IMPLEMENTED[scf.pot](scf)
    except KeyError:
        log.exception(f'No potential found for "{scf.pot}".')
        raise
    return pot


#: Map potential names with their respective implementation.
IMPLEMENTED = {
    'harmonic': harmonic,
    'coulomb': coulomb,
    'ge': ge,
    'gth': init_gth_loc
}



#!/usr/bin/env python3
"""eminus - A plane wave density functional theory code.

Minimal usage example to do a DFT calculation for helium::

   from eminus import Atoms, SCF
   atoms = Atoms('He', (0, 0, 0))
   SCF(atoms).run()
"""
from . import config
from .atoms import Atoms
from .dft import get_epsilon, get_psi
from .io import (
    read,
    read_cube,
    read_json,
    read_xyz,
    write,
    write_cube,
    write_json,
    write_pdb,
    write_xyz,
)
from .logger import log
from .scf import RSCF, SCF, USCF
from .version import __version__, info

__all__ = ['config', 'Atoms', 'get_epsilon', 'get_psi', 'info', 'log', 'read', 'read_cube',
           'read_json', 'read_xyz', 'RSCF', 'SCF', 'USCF', 'write', 'write_cube', 'write_json',
           'write_pdb', 'write_xyz', '__version__']


def demo():
    """Fast demo calculation for helium."""
    atoms = Atoms('He', (0, 0, 0), ecut=5)
    SCF(atoms).run()



#!/usr/bin/env python3
"""Perdew-Burke-Ernzerhof GGA correlation for solids and surfaces.

Reference: Phys. Rev. Lett. 102, 039902.
"""
from .gga_c_pbe import gga_c_pbe, gga_c_pbe_spin


def gga_c_pbe_sol(n, **kwargs):
    """Perdew-Burke-Ernzerhof solid parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label GGA_C_PBE_SOL and ID 133 in Libxc.

    Reference: Phys. Rev. Lett. 102, 039902.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE correlation energy density, potential, and vsigma.
    """
    return gga_c_pbe(n, beta=0.046, **kwargs)


def gga_c_pbe_sol_spin(n, zeta, **kwargs):
    """Perdew-Burke-Ernzerhof solid parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label GGA_C_PBE_SOL and ID 133 in Libxc.

    Reference: Phys. Rev. Lett. 102, 039902.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE correlation energy density, potential, and vsigma.
    """
    return gga_c_pbe_spin(n, zeta, beta=0.046, **kwargs)



#!/usr/bin/env python3
"""Interface to use Libxc functionals.

For a list of available functionals, see: https://tddft.org/programs/libxc/functionals

One can install pylibxc by compiling Libxc according to
https://tddft.org/programs/libxc/installation/#python-library

Alternatively, one can use the PySCF Libxc interface with::

    pip install eminus[libxc]
"""
import numpy as np
from scipy.linalg import norm

from .. import config
from ..logger import log


def libxc_functional(xc, n_spin, Nspin, dn_spin=None, tau=None):
    """Handle Libxc exchange-correlation functionals via pylibxc.

    Only LDA and GGA functionals can be used.

    Reference: SoftwareX 7, 1.

    Args:
        xc (str | int): Exchange or correlation identifier.
        n_spin (ndarray): Real-space electronic densities per spin channel.
        Nspin (int): Number of spin states.

    Keyword Args:
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        tau (ndarray): Real-space kinetic energy densities per spin channel.

    Returns:
        tuple[ndarray, ndarray, ndarray, ndarray]: Exch.-corr. energy density and potentials.
    """
    try:
        assert config.use_pylibxc
        from pylibxc import LibXCFunctional
    except (ImportError, AssertionError):
        return pyscf_functional(xc, n_spin, Nspin, dn_spin, tau)

    # Libxc functionals have one integer and one string identifier
    try:
        func = LibXCFunctional(int(xc), Nspin)
    except ValueError:
        func = LibXCFunctional(xc, Nspin)

    if dn_spin is None:
        # Libxc expects a 1d array, so reshape n_spin (same as n_spin.ravel(order='F'))
        out = func.compute({'rho': n_spin.T.ravel()})
    else:
        # The gradients have to be reshaped as well but also squared
        if Nspin == 1:
            dn2 = norm(dn_spin, axis=2)**2
        else:
            # For the spin-polarized case the gradients of spin-up and -down are mixed
            dn2 = np.vstack((norm(dn_spin[0], axis=1)**2, np.sum(dn_spin[0] * dn_spin[1], axis=1),
                            norm(dn_spin[1], axis=1)**2))
        if tau is None:
            out = func.compute({'rho': n_spin.T.ravel(), 'sigma': dn2.T.ravel()})
        else:
            out = func.compute({'rho': n_spin.T.ravel(), 'sigma': dn2.T.ravel(),
                                'tau': tau.T.ravel()})
    # zk is a column vector, flatten it to a 1d row vector
    exc = out['zk'].ravel()

    # vrho (and vsigma) is exactly transposed from what we need
    vxc = out['vrho'].T
    if dn_spin is not None:
        vsigma = np.atleast_2d(out['vsigma'].T)
        if tau is not None:
            vtau = out['vtau'].T
            return exc, vxc, vsigma, vtau
        return exc, vxc, vsigma, None
    return exc, vxc, None, None


def pyscf_functional(xc, n_spin, Nspin, dn_spin=None, tau=None):
    """Handle Libxc exchange-correlation functionals via PySCF.

    Only LDA and GGA functionals can be used.

    Reference: WIREs Comput. Mol. Sci. 8, e1340.

    Args:
        xc (str | int): Exchange or correlation identifier.
        n_spin (ndarray): Real-space electronic densities per spin channel.
        Nspin (int): Number of spin states.

    Keyword Args:
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        tau (ndarray): Real-space kinetic energy densities per spin channel.

    Returns:
        tuple[ndarray, ndarray, ndarray, ndarray]: Exch.-corr. energy density and potentials.
    """
    try:
        from pyscf.dft.libxc import eval_xc
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[libxc]".\n\n')
        raise

    if dn_spin is None:
        # For LDAs we only need the spin densities that are already in the needed shape
        rho = n_spin
    else:
        if tau is None:
            # For GGAs we have to append the density gradients
            # The input "density" rho is sorted as (n,grad_x n,grad_y n,grad_z n)
            if Nspin == 1:
                # For spin-paired systems we have to remove the spin indexing (the outermost shape)
                rho = np.vstack((n_spin[0], dn_spin[0].T))
            else:
                rho = np.array([np.vstack((n_spin[0], dn_spin[0].T)),
                                np.vstack((n_spin[1], dn_spin[1].T))])
        else:
            # For meta-GGAs we have to append the kinetic energy densities as well
            # The input "density" rho is sorted as (n,grad_x n,grad_y n,grad_z n,lapl,tau)
            # We do not support meta-GGAs with a dependency of the laplacian so set it to zero
            lapl = np.zeros(len(n_spin[0]))
            if Nspin == 1:
                rho = np.vstack((n_spin[0], dn_spin[0].T, lapl, tau[0]))
            else:
                rho = np.array([np.vstack((n_spin[0], dn_spin[0].T, lapl, tau[0])),
                                np.vstack((n_spin[1], dn_spin[1].T, lapl, tau[1]))])

    # Spin in PySCF is the number of unpaired electrons, not the number of spin channels
    exc, vxc, _, _ = eval_xc(xc, rho, spin=Nspin - 1)
    # The first entry of vxc is vrho
    # The second entry of the second entry is vsigma
    # The fourth entry of the second entry is vtau (the third would be vlapl)
    # These arrays are 1d for Nspin=1 and a 2d column array for Nspin=2, reshape them as needed
    if dn_spin is not None:
        if tau is not None:
            return exc, np.atleast_2d(vxc[0].T), np.atleast_2d(vxc[1].T), np.atleast_2d(vxc[3].T)
        return exc, np.atleast_2d(vxc[0].T), np.atleast_2d(vxc[1].T), None
    return exc, np.atleast_2d(vxc[0].T), None, None



#!/usr/bin/env python3
"""Consolidated configuration module."""
import numbers
import os
import pathlib
import sys
from typing import Callable, Union

from .logger import log


class ConfigClass:
    """Configuration class holding user specifiable variables.

    An instance of this class will be set as the same name as this module. This will effectively
    make this module a singleton data class.
    """
    def __init__(self):
        """Initialize the ConfigClass object."""
        self.use_torch = True     # Use the faster Torch FFTs if available
        self.use_gpu = False      # Disable GPU by default, since it is slower in my tests
        self.use_pylibxc = True   # Use Libxc over PySCF if available, since it is faster
        self.threads = None       # Read threads from environment variables by default
        self.verbose = 'INFO'     # Only display warnings (and worse) by default

    # ### Class properties ###

    @property
    def use_torch(self):
        """Weather to use Torch or SciPy if Torch is installed."""
        # Add the logic in the getter method so it does not run on initialization since importing
        # Torch is rather slow
        if self._use_torch:
            try:
                import torch  # noqa: F401
                return True
            except ImportError:
                pass
        return False

    @use_torch.setter
    def use_torch(self, value):
        self._use_torch = value

    @property
    def use_gpu(self):
        """Weather to use Torch on the GPU if available."""
        # Only use GPU if Torch is available
        if self.use_torch and self._use_gpu:
            import torch
            return torch.cuda.is_available()
        return False

    @use_gpu.setter
    def use_gpu(self, value):
        self._use_gpu = value

    @property
    def use_pylibxc(self):
        """Weather to use pylibxc or PySCF for functionals if both are installed."""
        if self._use_pylibxc:
            try:
                import pylibxc  # noqa: F401
                return True
            except ImportError:
                pass
        return False

    @use_pylibxc.setter
    def use_pylibxc(self, value):
        self._use_pylibxc = value

    @property
    def threads(self):
        """Number of threads used in FFT calculations."""
        if self._threads is None:
            try:
                if self.use_torch:
                    import torch
                    return torch.get_num_threads()
                # Read the OMP threads for the default operators
                return int(os.environ['OMP_NUM_THREADS'])
            except KeyError:
                return None
        return int(self._threads)

    @threads.setter
    def threads(self, value):
        self._threads = value
        if isinstance(value, numbers.Integral):
            if self.use_torch:
                import torch
                return torch.set_num_threads(value)
            os.environ['OMP_NUM_THREADS'] = str(value)
        return None

    @property
    def verbose(self):
        """Logger verbosity level."""
        return log.verbose

    @verbose.setter
    def verbose(self, value):
        # Logic in setter to run it on initialization
        log.verbose = value

    # ### Class methods ###

    def info(self):
        """Print configuration and performance information."""
        print('--- Configuration infos ---')
        print(f'Global verbosity : {self.verbose}')
        # Only print if PySCF or pylibxc is installed
        if not self.use_pylibxc:
            try:
                import pyscf  # noqa: F401
                print('Libxc backend    : PySCF')
            except ImportError:
                pass
        else:
            print('Libxc backend    : pylibxc')

        print('\n--- Performance infos ---')
        print(f'FFT backend : {"Torch" if self.use_torch else "SciPy"}')
        print(f'FFT device  : {"GPU" if self.use_gpu else "CPU"}')
        # Do not print threading information when using GPU
        if self.use_gpu:
            return
        # Check FFT threads
        if self.threads is None:
            print('FFT threads : 1\n'
                  'INFO: No OMP_NUM_THREADS environment variable was found.\nTo improve '
                  'performance, add "export OMP_NUM_THREADS=n" to your ".bashrc".\nMake sure to '
                  'replace "n", typically with the number of cores your CPU.\nTemporarily, you can '
                  'set them in your Python environment with "eminus.config.threads=n".')
        else:
            print(f'FFT threads : {self.threads}')


# Add type hints for all properties and methods of the ConfigClass to the module
# This allows type checkers to see that the module has said attribute
use_torch: bool
use_gpu: bool
use_pylibxc: bool
threads: int
verbose: Union[int, str]
info: Callable[[], None]

# Do not initialize the class when Sphinx is running
# Since we set the class instance to the module name Sphinx will only document the main docstring of
# the class without the properties
if 'sphinx-build' not in pathlib.Path(sys.argv[0]).name:
    sys.modules[__name__] = ConfigClass()  # type: ignore[assignment]



#!/usr/bin/env python3
"""PBE GTH pseudopotential files.

Reference: Theor. Chem. Acc. 114, 145.
"""

if __name__ == '__main__':
    import inspect
    import pathlib
    import shutil
    import urllib.request
    import zipfile

    psp_path = pathlib.Path(inspect.stack()[0][1]).parent
    file = 'master.zip'
    # Download files
    url = f'https://github.com/cp2k/cp2k-data/archive/refs/heads/{file}'
    urllib.request.urlretrieve(url, file)  # noqa: S310
    # Unpack files
    with zipfile.ZipFile(file, 'r') as fzip:
        fzip.extractall()
    # Move files
    gth_path = psp_path.joinpath('cp2k-data-master/potentials/Goedecker/cp2k/pbe')
    for f in gth_path.iterdir():
        shutil.move(str(f), psp_path.joinpath(f.name))
    # Cleanup
    psp_path.joinpath(file).unlink()
    shutil.rmtree('cp2k-data-master')



#!/usr/bin/env python3
"""Utilities to use Goedecker, Teter, and Hutter pseudopotentials."""
import numpy as np

from .io import read_gth
from .utils import Ylm_real


class GTH:
    """GTH object that holds non-local projectors and parameters for all atoms in the SCF object.

    Keyword Args:
        scf: SCF object.
    """
    def __init__(self, scf=None):
        """Initialize the GTH object."""
        # Allow creating an empty instance (used when loading GTH objects from JSON files)
        if scf is not None:
            atoms = scf.atoms

            # Set up a dictionary with all GTH parameters
            self.GTH = {}  #: Dictionary with GTH parameters per atom type.
            for ia in range(atoms.Natoms):
                # Skip if the atom is already in the dictionary
                if atoms.atom[ia] in self.GTH:
                    continue
                self.GTH[atoms.atom[ia]] = read_gth(atoms.atom[ia], atoms.Z[ia], psp_path=scf.psp)

            # Initialize the non-local potential
            NbetaNL, prj2beta, betaNL = init_gth_nonloc(atoms, self)
            self.NbetaNL = NbetaNL    #: Number of projector functions for the non-local potential.
            self.prj2beta = prj2beta  #: Index matrix to map to the correct projector function.
            self.betaNL = betaNL      #: Atomic-centered projector functions.

    def __getitem__(self, key):
        """Allow accessing the GTH parameters of an atom by indexing the GTH object."""
        return self.GTH[key]

    def __repr__(self):
        """Print a short overview over the values stored in the GTH object."""
        return f'NbetaNL: {self.NbetaNL}\nGTH values for: {", ".join(list(self.GTH))}'


def init_gth_loc(scf):
    """Initialize parameters to calculate local contributions of GTH pseudopotentials.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        scf: SCF object.

    Returns:
        ndarray: Local GTH potential contribution.
    """
    atoms = scf.atoms
    species = set(atoms.atom)
    omega = 1  # Normally this would be det(atoms.R), but Arias notation is off by this factor

    Vloc = np.zeros_like(atoms.G2)
    for isp in species:
        psp = scf.gth[isp]
        rloc = psp['rloc']
        Zion = psp['Zion']
        c1 = psp['cloc'][0]
        c2 = psp['cloc'][1]
        c3 = psp['cloc'][2]
        c4 = psp['cloc'][3]

        rlocG2 = atoms.G2 * rloc**2
        rlocG22 = rlocG2**2
        exprlocG2 = np.exp(-0.5 * rlocG2)
        # Ignore the division by zero for the first elements
        # One could do some proper indexing with [1:] but indexing is slow
        with np.errstate(divide='ignore', invalid='ignore'):
            Vsp = -4 * np.pi * Zion / omega * exprlocG2 / atoms.G2 + \
                np.sqrt((2 * np.pi)**3) * rloc**3 / omega * exprlocG2 * \
                (c1 + c2 * (3 - rlocG2) + c3 * (15 - 10 * rlocG2 + rlocG22) +
                 c4 * (105 - 105 * rlocG2 + 21 * rlocG22 - rlocG2**3))
        # Special case for G=(0,0,0), same as in QE
        Vsp[0] = 2 * np.pi * Zion * rloc**2 + \
            (2 * np.pi)**1.5 * rloc**3 * (c1 + 3 * c2 + 15 * c3 + 105 * c4)

        # Sum up the structure factor for every species
        Sf = np.zeros(len(atoms.Sf[0]), dtype=complex)
        for ia in range(atoms.Natoms):
            if atoms.atom[ia] == isp:
                Sf += atoms.Sf[ia]
        Vloc += np.real(atoms.J(Vsp * Sf))
    return Vloc


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/PsPotNL.jl
def init_gth_nonloc(atoms, gth):
    """Initialize parameters to calculate non-local contributions of GTH pseudopotentials.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        atoms: Atoms object.
        gth: GTH object.

    Returns:
        tuple[int, ndarray, ndarray]: NbetaNL, prj2beta, and betaNL.
    """
    prj2beta = np.zeros((3, atoms.Natoms, 4, 7), dtype=int)
    prj2beta += -1  # Set to an invalid index

    NbetaNL = 0
    for ia in range(atoms.Natoms):
        psp = gth[atoms.atom[ia]]
        for l in range(psp['lmax']):
            for m in range(-l, l + 1):
                for iprj in range(psp['Nproj_l'][l]):
                    NbetaNL += 1
                    prj2beta[iprj, ia, l, m + psp['lmax'] - 1] = NbetaNL

    g = atoms.G[atoms.active]  # Simplified, would normally be G+k
    Gm = np.sqrt(atoms.G2c)

    ibeta = 0
    betaNL = np.empty((len(atoms.G2c), NbetaNL), dtype=complex)
    for ia in range(atoms.Natoms):
        # It is very important to transform the structure factor to make both notations compatible
        Sf = atoms.Idag(atoms.J(atoms.Sf[ia]))
        psp = gth[atoms.atom[ia]]
        for l in range(psp['lmax']):
            for m in range(-l, l + 1):
                for iprj in range(psp['Nproj_l'][l]):
                    betaNL[:, ibeta] = (-1j)**l * Ylm_real(l, m, g) * \
                        eval_proj_G(psp, l, iprj + 1, Gm, atoms.Omega) * Sf
                    ibeta += 1
    return NbetaNL, prj2beta, betaNL


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/op_V_Ps_nloc.jl
def calc_Vnonloc(scf, spin, W):
    """Calculate the non-local pseudopotential, applied on the basis functions W.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        scf: SCF object.
        spin (int): Spin variable to track weather to do the calculation for spin up or down.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: Non-local GTH potential contribution.
    """
    atoms = scf.atoms

    Vpsi = np.zeros_like(W[spin], dtype=complex)
    if scf.pot == 'gth' and scf.gth.NbetaNL > 0:  # Only calculate the non-local part if necessary
        betaNL_psi = (W[spin].conj().T @ scf.gth.betaNL).conj()

        for ia in range(atoms.Natoms):
            psp = scf.gth[atoms.atom[ia]]
            for l in range(psp['lmax']):
                for m in range(-l, l + 1):
                    for iprj in range(psp['Nproj_l'][l]):
                        ibeta = scf.gth.prj2beta[iprj, ia, l, m + psp['lmax'] - 1] - 1
                        for jprj in range(psp['Nproj_l'][l]):
                            jbeta = scf.gth.prj2beta[jprj, ia, l, m + psp['lmax'] - 1] - 1
                            hij = psp['h'][l, iprj, jprj]
                            Vpsi += hij * betaNL_psi[:, jbeta] * scf.gth.betaNL[:, ibeta, None]
    # We have to multiply with the cell volume, because of different orthogonalization methods
    return Vpsi * atoms.Omega


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/PsPot_GTH.jl
def eval_proj_G(psp, l, iprj, Gm, Omega):
    """Evaluate GTH projector functions in G-space.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        psp (dict): GTH parameters.
        l (int): Angular momentum number.
        iprj (int): Nproj_l index.
        Gm (ndarray): Magnitude of G-vectors.
        Omega (float): Unit cell volume.

    Returns:
        ndarray: GTH projector.
    """
    rrl = psp['rp'][l]
    Gr2 = (Gm * rrl)**2

    prefactor = 4 * np.pi**(5 / 4) * np.sqrt(2**(l + 1) * rrl**(2 * l + 3) / Omega)
    Vprj = prefactor * np.exp(-0.5 * Gr2)

    if l == 0:  # s-channel
        if iprj == 1:
            return Vprj
        if iprj == 2:
            return 2 / np.sqrt(15) * (3 - Gr2) * Vprj
        if iprj == 3:
            return (4 / 3) / np.sqrt(105) * (15 - 10 * Gr2 + Gr2**2) * Vprj
    elif l == 1:  # p-channel
        if iprj == 1:
            return 1 / np.sqrt(3) * Gm * Vprj
        if iprj == 2:
            return 2 / np.sqrt(105) * Gm * (5 - Gr2) * Vprj
        if iprj == 3:
            return (4 / 3) / np.sqrt(1155) * Gm * (35 - 14 * Gr2 + Gr2**2) * Vprj
    elif l == 2:  # d-channel
        if iprj == 1:
            return 1 / np.sqrt(15) * Gm**2 * Vprj
        if iprj == 2:
            return (2 / 3) / np.sqrt(105) * Gm**2 * (7 - Gr2) * Vprj
    elif l == 3:  # f-channel
        # Only one projector
        return 1 / np.sqrt(105) * Gm**3 * Vprj

    ValueError(f'No projector found for l={l}.')
    return None



#!/usr/bin/env python3
"""Chachiyo LDA correlation.

Reference: J. Chem. Phys. 145, 021101.
"""
import numpy as np


def lda_c_chachiyo(n, **kwargs):
    """Chachiyo parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_CHACHIYO and ID 287 in Libxc.

    Reference: J. Chem. Phys. 145, 021101.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: Chachiyo correlation energy density and potential.
    """
    a = -0.01554535  # (np.log(2) - 1) / (2 * np.pi**2)
    b = 20.4562557

    rs = (3 / (4 * np.pi * n))**(1 / 3)
    rs2 = rs**2
    ecinner = 1 + b / rs + b / rs2

    ec = a * np.log(ecinner)

    vc = ec + a * b * (2 + rs) / (3 * (b + b * rs + rs2))
    return ec, np.array([vc]), None


def chachiyo_scaling(zeta):
    """Weighting factor between the paramagnetic and the ferromagnetic case.

    Reference: J. Chem. Phys. 145, 021101.

    Args:
        zeta (ndarray): Relative spin polarization.

    Returns:
        tuple[ndarray, ndarray]: Weighting factor and its derivative.
    """
    fzeta = ((1 + zeta)**(4 / 3) + (1 - zeta)**(4 / 3) - 2) / (2 * (2**(1 / 3) - 1))

    dfdzeta = (2 * (1 - zeta)**(1 / 3) - 2 * (1 + zeta)**(1 / 3)) / (3 - 3 * 2**(1 / 3))
    return fzeta, dfdzeta


def lda_c_chachiyo_spin(n, zeta, weight_function=chachiyo_scaling, **kwargs):
    """Chachiyo parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_CHACHIYO and ID 287 in Libxc.

    Reference: J. Chem. Phys. 145, 021101.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        weight_function (Callable): Functional function.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: Chachiyo correlation energy density and potential.
    """
    a0 = -0.01554535   # (np.log(2) - 1) / (2 * np.pi**2)
    a1 = -0.007772675  # (np.log(2) - 1) / (4 * np.pi**2)
    b0 = 20.4562557
    b1 = 27.4203609

    rs = (3 / (4 * np.pi * n))**(1 / 3)
    rs2 = rs**2

    fzeta, dfdzeta = weight_function(zeta)

    ec0inner = 1 + b0 / rs + b0 / rs2
    ec1inner = 1 + b1 / rs + b1 / rs2
    ec0 = a0 * np.log(ec0inner)
    ec1 = a1 * np.log(ec1inner)

    ec = ec0 + (ec1 - ec0) * fzeta

    factor = -1 / rs2 - 2 / rs**3
    dec0drs = a0 / ec0inner * b0 * factor
    dec1drs = a1 / ec1inner * b1 * factor
    decdrs = dec0drs + (dec1drs - dec0drs) * fzeta
    prefactor = ec - rs / 3 * decdrs
    decdf = (ec1 - ec0) * dfdzeta

    vcup = prefactor + decdf * (1 - zeta)
    vcdw = prefactor - decdf * (1 + zeta)
    return ec, np.array([vcup, vcdw]), None



#!/usr/bin/env python3
"""Implementation of operators using Torch FFT functions.

For more details see :mod:`~eminus.operators`.

For installation instructions see https://pytorch.org/get-started/locally.

This implementation is focused on speed, rather than readability since these operators need the
most time in most calculations. Notable differences to the default operators are:

* Use Torch FFTs (we need to cast atoms.s to tuples for this)
* No handle_spin_gracefully decorator to reduce function calls and casts
* Call properties directly using their private attribute
* Explicitly implement Idag and Jdag for one function call less and direct normalization
* Optional GPU calculations

In my tests, the overhead to move the arrays to the GPU and back is not worth it, so it is disabled
by default.

Reference: Adv. Neural. Inf. Process Syst. 32, 8024.
"""
import numpy as np

from .. import config


def I(atoms, W):
    """Backwards transformation from reciprocal space to real-space.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    """
    import torch
    n = atoms._Ns
    s = tuple(atoms._s)

    if W.ndim < 3:
        if len(W) == len(atoms._G2):
            Wfft = W
        else:
            if W.ndim == 1:
                Wfft = np.zeros(n, dtype=W.dtype)
            else:
                Wfft = np.zeros((n, atoms.occ._Nstate), dtype=W.dtype)
            Wfft[atoms._active] = W
    else:
        if W.shape[1] == len(atoms._G2):
            Wfft = W
        else:
            Wfft = np.zeros((atoms.occ._Nspin, n, atoms.occ._Nstate), dtype=W.dtype)
            Wfft[:, atoms._active[0]] = W

    Wfft = torch.from_numpy(Wfft)
    if config.use_gpu:
        Wfft = Wfft.cuda()

    if W.ndim == 1:
        Wfft = Wfft.view(s)
        Finv = torch.fft.ifftn(Wfft, s=s, norm='forward').view(n)
    elif W.ndim == 2:
        Wfft = Wfft.view(s + (atoms.occ._Nstate,))
        Finv = torch.fft.ifftn(Wfft, s=s, norm='forward', dim=(0, 1, 2)).view(n, atoms.occ._Nstate)
    else:
        Wfft = Wfft.view((atoms.occ._Nspin,) + s + (atoms.occ._Nstate,))
        Finv = torch.fft.ifftn(Wfft, s=s, norm='forward', dim=(1, 2, 3)).view(atoms.occ._Nspin, n,
                                                                              atoms.occ._Nstate)
    return Finv.detach().cpu().numpy()


def J(atoms, W, full=True):
    """Forward transformation from real-space to reciprocal space.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        full (bool): Wether to transform in the full or in the active space.

    Returns:
        ndarray: The operator applied on W.
    """
    import torch
    n = atoms._Ns
    s = tuple(atoms._s)

    Wfft = torch.from_numpy(W)
    if config.use_gpu:
        Wfft = Wfft.cuda()

    if W.ndim == 1:
        Wfft = Wfft.view(s)
        F = torch.fft.fftn(Wfft, s=s, norm='forward').view(n)
    elif W.ndim == 2:
        Wfft = Wfft.view(s + (atoms.occ._Nstate,))
        F = torch.fft.fftn(Wfft, s=s, norm='forward', dim=(0, 1, 2)).view(n, atoms.occ._Nstate)
    else:
        Wfft = Wfft.view((atoms.occ._Nspin,) + s + (atoms.occ._Nstate,))
        F = torch.fft.fftn(Wfft, s=s, norm='forward', dim=(1, 2, 3)).view(atoms.occ._Nspin, n,
                                                                          atoms.occ._Nstate)
    F = F.detach().cpu().numpy()
    if not full:
        if F.ndim < 3:
            return F[atoms._active]
        return F[:, atoms._active[0]]
    return F


def Idag(atoms, W, full=False):
    """Conjugated backwards transformation from real-space to reciprocal space.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        full (bool): Wether to transform in the full or in the active space.

    Returns:
        ndarray: The operator applied on W.
    """
    import torch
    n = atoms._Ns
    s = tuple(atoms._s)

    Wfft = torch.from_numpy(W)
    if config.use_gpu:
        Wfft = Wfft.cuda()

    if W.ndim == 1:
        Wfft = Wfft.view(s)
        F = torch.fft.fftn(Wfft, s=s, norm='backward').view(n)
    elif W.ndim == 2:
        Wfft = Wfft.view(s + (atoms.occ._Nstate,))
        F = torch.fft.fftn(Wfft, s=s, norm='backward', dim=(0, 1, 2)).view(n, atoms.occ._Nstate)
    else:
        Wfft = Wfft.view((atoms.occ._Nspin,) + s + (atoms.occ._Nstate,))
        F = torch.fft.fftn(Wfft, s=s, norm='backward', dim=(1, 2, 3)).view(atoms.occ._Nspin, n,
                                                                           atoms.occ._Nstate)
    F = F.detach().cpu().numpy()
    if not full:
        if F.ndim < 3:
            return F[atoms._active]
        return F[:, atoms._active[0]]
    return F


def Jdag(atoms, W):
    """Conjugated forward transformation from reciprocal space to real-space.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    """
    import torch
    n = atoms._Ns
    s = tuple(atoms._s)

    if W.ndim < 3:
        if len(W) == len(atoms._G2):
            Wfft = W
        else:
            if W.ndim == 1:
                Wfft = np.zeros(n, dtype=W.dtype)
            else:
                Wfft = np.zeros((n, atoms.occ._Nstate), dtype=W.dtype)
            Wfft[atoms._active] = W
    else:
        if W.shape[1] == len(atoms._G2):
            Wfft = W
        else:
            Wfft = np.zeros((atoms.occ._Nspin, n, atoms.occ._Nstate), dtype=W.dtype)
            Wfft[:, atoms._active[0]] = W

    Wfft = torch.from_numpy(Wfft)
    if config.use_gpu:
        Wfft = Wfft.cuda()

    if W.ndim == 1:
        Wfft = Wfft.view(s)
        Finv = torch.fft.ifftn(Wfft, s=s, norm='backward').view(n)
    elif W.ndim == 2:
        Wfft = Wfft.view(s + (atoms.occ._Nstate,))
        Finv = torch.fft.ifftn(Wfft, s=s, norm='backward', dim=(0, 1, 2)).view(n, atoms.occ._Nstate)
    else:
        Wfft = Wfft.view((atoms.occ._Nspin,) + s + (atoms.occ._Nstate,))
        Finv = torch.fft.ifftn(Wfft, s=s, norm='backward', dim=(1, 2, 3)).view(atoms.occ._Nspin, n,
                                                                               atoms.occ._Nstate)
    return Finv.detach().cpu().numpy()



#!/usr/bin/env python3
"""Minimization algorithms."""
import logging

import numpy as np

from .dft import get_grad, get_n_spin, get_n_total, orth, solve_poisson
from .energies import get_E
from .gga import get_grad_field, get_tau
from .logger import name
from .utils import dotprod
from .xc import get_xc


def scf_step(scf):
    """Perform one SCF step for a DFT calculation.

    Calculating intermediate results speeds up the energy and gradient calculation.
    This function is similar to H_precompute but will set all variables and energies in the SCF
    class and returns the total energy.

    Args:
        scf: SCF object.

    Returns:
        float: Total energy.
    """
    atoms = scf.atoms
    scf.Y = orth(atoms, scf.W)
    scf.n_spin = get_n_spin(atoms, scf.Y)
    scf.n = get_n_total(atoms, scf.Y, scf.n_spin)
    if 'gga' in scf.xc_type:
        scf.dn_spin = get_grad_field(atoms, scf.n_spin)
    if scf.xc_type == 'meta-gga':
        scf.tau = get_tau(atoms, scf.Y)
    scf.phi = solve_poisson(atoms, scf.n)
    scf.exc, scf.vxc, scf.vsigma, scf.vtau = get_xc(scf.xc, scf.n_spin, atoms.occ.Nspin,
                                                    scf.dn_spin, scf.tau)
    scf._precomputed = {'dn_spin': scf.dn_spin, 'phi': scf.phi, 'vxc': scf.vxc,
                        'vsigma': scf.vsigma, 'vtau': scf.vtau}
    return get_E(scf)


def check_convergence(scf, method, Elist, linmin=None, cg=None, norm_g=None):
    """Check the energies for every SCF cycle and handle the output.

    Args:
        scf: SCF object.
        method (string): Minimization method.
        Elist (list): Total energies per SCF step.

    Keyword Args:
        linmin (ndarray): Cosine between previous search direction and current gradient.
        cg (ndarray): Conjugate-gradient orthogonality.
        norm_g (ndarray): Gradient norm.

    Returns:
        bool: Convergence condition.
    """
    iteration = len(Elist)

    # Print all the data
    print_scf_step(scf, method, Elist, linmin, cg, norm_g)

    if iteration > 1:
        # Check for convergence
        if scf.gradtol is None or norm_g is None:
            if abs(Elist[-2] - Elist[-1]) < scf.etol:
                scf.is_converged = True
                return True
        # If a gradient tolerance has been set we also check norm_g for convergence
        else:
            if abs(Elist[-2] - Elist[-1]) < scf.etol and (norm_g < scf.gradtol).all():
                scf.is_converged = True
                return True
        # Check if the current energy is higher than the last two values
        if (np.asarray(Elist[-3:-1]) < Elist[-1]).all():
            scf.log.warning('Total energy is not decreasing.')
    return False


def print_scf_step(scf, method, Elist, linmin, cg, norm_g):
    """Print the data of one SCF step and the header at the beginning.

    Args:
        scf: SCF object.
        method (string): Minimization method.
        Elist (list): Total energies per SCF step.
        linmin (ndarray): Cosine between previous search direction and current gradient.
        cg (ndarray): Conjugate-gradient orthogonality.
        norm_g (ndarray): Gradient norm.
    """
    iteration = len(Elist)

    # Print a column header at the beginning
    # The ljust values just have been chosen such that the output looks decent
    if iteration == 1:
        header = 'Method'.ljust(8)
        header += 'Iteration'.ljust(11)
        header += 'Etot [Eh]'.ljust(13)
        header += 'dEtot [Eh]'.ljust(13)
        # Print the gradient norm for cg methods
        if method not in ('sd', 'lm', 'pclm'):
            header += '|Gradient|'.ljust(10 * scf.atoms.occ.Nspin + 3)
        # Print extra debugging information if available
        if scf.log.level <= logging.DEBUG:
            if method != 'sd':
                header += 'linmin-test'.ljust(10 * scf.atoms.occ.Nspin + 3)
            if method not in ('sd', 'lm', 'pclm'):
                header += 'cg-test'.ljust(10 * scf.atoms.occ.Nspin + 3)
            scf.log.debug(header)
        else:
            scf.log.info(header)

    # Print the information for every cycle
    # Context manager for printing norm_g, linmin, and cg
    with np.printoptions(formatter={'float': '{:+0.2e}'.format}):
        info = f'{method:<8}{iteration:>8}   {scf.energies.Etot:<+13,.6f}'
        # In the first step we do not have all information yet
        if iteration > 1:
            info += f'{Elist[-2] - Elist[-1]:<+13,.4e}'
            if norm_g is not None:
                info += str(norm_g).ljust(10 * scf.atoms.occ.Nspin + 3)
            if scf.log.level <= logging.DEBUG:
                if method != 'sd':
                    info += str(linmin).ljust(10 * scf.atoms.occ.Nspin + 3)
                if method not in ('sd', 'lm', 'pclm'):
                    info += str(cg).ljust(10 * scf.atoms.occ.Nspin + 3)
    if scf.log.level <= logging.DEBUG:
        scf.log.debug(info)
    else:
        scf.log.info(info)


def linmin_test(g, d):
    """Do the line minimization test.

    Calculate the cosine of the angle between g and d.

    Reference: https://trond.hjorteland.com/thesis/node26.html

    Args:
        g (ndarray): Current gradient.
        d (ndarray): Previous search direction.

    Returns:
        float: Linmin angle.
    """
    # cos = A B / |A| |B|
    return dotprod(g, d) / np.sqrt(dotprod(g, g) * dotprod(d, d))


def cg_test(atoms, g, g_old, precondition=True):
    """Test the gradient-orthogonality theorem, i.e., g and g_old should be orthogonal.

    Calculate the cosine of the angle between g and g_old. For an angle of 90 degree the cosine goes
    to zero.

    Reference: https://math.uci.edu/~chenlong/CAMtips/CG.html

    Args:
        atoms: Atoms object.
        g (ndarray): Current gradient.
        g_old (ndarray): Previous gradient.

    Keyword Args:
        precondition (bool): Weather to use a preconditioner.

    Returns:
        float: CG angle.
    """
    if precondition:
        Kg, Kg_old = atoms.K(g), atoms.K(g_old)
    else:
        Kg, Kg_old = g, g_old
    # cos = A B / |A| |B|
    return dotprod(g, Kg_old) / np.sqrt(dotprod(g, Kg) * dotprod(g_old, Kg_old))


def cg_method(scf, cgform, g, g_old, d_old, precondition=True):
    """Do different variants of the conjugate gradient method.

    Reference: https://indrag49.github.io/Numerical-Optimization/conjugate-gradient-methods-1.html

    Args:
        scf: SCF object.
        cgform (int): Conjugate gradient form.
        g (ndarray): Current gradient.
        g_old (ndarray): Previous gradient.
        d_old (ndarray): Previous search direction.

    Keyword Args:
        precondition (bool): Weather to use a preconditioner.

    Returns:
        tuple[float, float]: Conjugate scalar and gradient norm.
    """
    atoms = scf.atoms

    if precondition:
        Kg, Kg_old = atoms.K(g), atoms.K(g_old)
    else:
        Kg, Kg_old = g, g_old
    norm_g = dotprod(g, Kg)

    if cgform == 1:    # Fletcher-Reeves
        return norm_g / dotprod(g_old, Kg_old), norm_g
    if cgform == 2:  # Polak-Ribiere
        return dotprod(g - g_old, Kg) / dotprod(g_old, Kg_old), norm_g
    if cgform == 3:  # Hestenes-Stiefel
        return dotprod(g - g_old, Kg) / dotprod(g - g_old, d_old), norm_g
    if cgform == 4:  # Dai-Yuan
        return norm_g / dotprod(g - g_old, d_old), norm_g
    ValueError(f'No cgform found for "{cgform}".')
    return None


@name('steepest descent minimization')
def sd(scf, Nit, cost=scf_step, grad=get_grad, condition=check_convergence, betat=3e-5, **kwargs):
    """Steepest descent minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        **kwargs: Throwaway arguments.

    Returns:
        list: Total energies per SCF cycle.
    """
    atoms = scf.atoms
    costs = []

    for _ in range(Nit):
        c = cost(scf)
        costs.append(c)
        if condition(scf, 'sd', costs):
            break
        for spin in range(atoms.occ.Nspin):
            g = grad(scf, spin, scf.W, **scf._precomputed)
            scf.W[spin] = scf.W[spin] - betat * g
    return costs


@name('preconditioned line minimization')
def pclm(scf, Nit, cost=scf_step, grad=get_grad, condition=check_convergence, betat=3e-5,
         precondition=True, **kwargs):
    """Preconditioned line minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        precondition (bool): Weather to use a preconditioner.
        **kwargs: Throwaway arguments.

    Returns:
        list: Total energies per SCF cycle.
    """
    atoms = scf.atoms
    costs = []

    if precondition:
        method = 'pclm'
    else:
        method = 'lm'

    # Scalars that need to be saved for each spin
    linmin = np.empty(atoms.occ.Nspin)
    beta = np.empty((atoms.occ.Nspin, 1, 1))
    # Search direction that needs to be saved for each spin
    d = np.empty_like(scf.W, dtype=complex)

    for _ in range(Nit):
        for spin in range(atoms.occ.Nspin):
            g = grad(scf, spin, scf.W, **scf._precomputed)
            # Calculate linmin each spin separately
            if scf.log.level <= logging.DEBUG and Nit > 0:
                linmin[spin] = linmin_test(g, d[spin])
            if precondition:
                d[spin] = -atoms.K(g)
            else:
                d[spin] = -g
            gt = grad(scf, spin, scf.W + betat * d[spin])
            beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])

        scf.W = scf.W + beta * d
        c = cost(scf)
        costs.append(c)
        if condition(scf, method, costs, linmin):
            break
    return costs


@name('line minimization')
def lm(scf, Nit, cost=scf_step, grad=get_grad, condition=check_convergence, betat=3e-5, **kwargs):
    """Line minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        **kwargs: Throwaway arguments.

    Returns:
        list: Total energies per SCF cycle.
    """
    return pclm(scf, Nit, cost, grad, condition, betat, precondition=False)


@name('preconditioned conjugate-gradient minimization')
def pccg(scf, Nit, cost=scf_step, grad=get_grad, condition=check_convergence, betat=3e-5, cgform=1,
         precondition=True):
    """Preconditioned conjugate-gradient minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        cgform (int): Conjugate gradient form.
        precondition (bool): Weather to use a preconditioner.

    Returns:
        list: Total energies per SCF cycle.
    """
    atoms = scf.atoms
    costs = []

    if precondition:
        method = 'pccg'
    else:
        method = 'cg'

    # Scalars that need to be saved for each spin
    linmin = np.empty(atoms.occ.Nspin)
    cg = np.empty(atoms.occ.Nspin)
    beta = np.empty((atoms.occ.Nspin, 1, 1))
    norm_g = np.empty(atoms.occ.Nspin)
    # Gradients that need to be saved for each spin
    d = np.empty_like(scf.W, dtype=complex)
    d_old = np.empty_like(scf.W, dtype=complex)
    g_old = np.empty_like(scf.W, dtype=complex)

    # Do the first step without the linmin and cg tests, and without the cg_method
    for spin in range(atoms.occ.Nspin):
        g = grad(scf, spin, scf.W)
        if precondition:
            d[spin] = -atoms.K(g)
        else:
            d[spin] = -g
        gt = grad(scf, spin, scf.W + betat * d[spin])
        beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
        g_old[spin], d_old[spin] = g, d[spin]

    # Update wave functions after calculating the gradients for each spin
    scf.W = scf.W + beta * d
    c = cost(scf)
    costs.append(c)
    condition(scf, method, costs)

    for _ in range(1, Nit):
        for spin in range(atoms.occ.Nspin):
            g = grad(scf, spin, scf.W, **scf._precomputed)
            # Calculate linmin and cg for each spin separately
            if scf.log.level <= logging.DEBUG:
                linmin[spin] = linmin_test(g, d[spin])
                cg[spin] = cg_test(atoms, g, g_old[spin], precondition)
            beta[spin], norm_g[spin] = cg_method(scf, cgform, g, g_old[spin], d_old[spin],
                                                 precondition)
            if precondition:
                d[spin] = -atoms.K(g) + beta[spin] * d_old[spin]
            else:
                d[spin] = -g + beta[spin] * d_old[spin]
            gt = grad(scf, spin, scf.W + betat * d[spin])
            beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
            g_old[spin], d_old[spin] = g, d[spin]

        scf.W = scf.W + beta * d
        c = cost(scf)
        costs.append(c)
        if condition(scf, method, costs, linmin, cg, norm_g):
            break
    return costs


@name('conjugate-gradient minimization')
def cg(scf, Nit, cost=scf_step, grad=get_grad, condition=check_convergence, betat=3e-5, cgform=1):
    """Conjugate-gradient minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        cgform (int): Conjugate gradient form.

    Returns:
        list: Total energies per SCF cycle.
    """
    return pccg(scf, Nit, cost, grad, condition, betat, cgform, precondition=False)


@name('auto minimization')
def auto(scf, Nit, cost=scf_step, grad=get_grad, condition=check_convergence, betat=3e-5, cgform=1):
    """Automatic preconditioned conjugate-gradient minimization algorithm.

    This function chooses an sd step over the pccg step if the energy goes up.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        cgform (int): Conjugate gradient form.

    Returns:
        list: Total energies per SCF cycle.
    """
    atoms = scf.atoms
    costs = []

    # Scalars that need to be saved for each spin
    linmin = np.empty(atoms.occ.Nspin)
    cg = np.empty(atoms.occ.Nspin)
    beta = np.empty((atoms.occ.Nspin, 1, 1))
    norm_g = np.empty(atoms.occ.Nspin)
    # Gradients that need to be saved for each spin
    g = np.empty_like(scf.W, dtype=complex)
    d = np.empty_like(scf.W, dtype=complex)
    d_old = np.empty_like(scf.W, dtype=complex)
    g_old = np.empty_like(scf.W, dtype=complex)

    # Do the first step without the linmin and cg tests, and without the cg_method
    for spin in range(atoms.occ.Nspin):
        g[spin] = grad(scf, spin, scf.W, **scf._precomputed)
        d[spin] = -atoms.K(g[spin])
        gt = grad(scf, spin, scf.W + betat * d[spin])
        beta[spin] = betat * dotprod(g[spin], d[spin]) / dotprod(g[spin] - gt, d[spin])
        g_old[spin], d_old[spin] = g[spin], d[spin]

    # Update wave functions after calculating the gradients for each spin, save the wave function
    W_old = np.copy(scf.W)
    scf.W = scf.W + beta * d
    c = cost(scf)
    costs.append(c)
    if condition(scf, 'pccg', costs):
        return costs

    for _ in range(1, Nit):
        for spin in range(atoms.occ.Nspin):
            g[spin] = grad(scf, spin, scf.W, **scf._precomputed)
            # Calculate linmin and cg for each spin separately
            if scf.log.level <= logging.DEBUG:
                linmin[spin] = linmin_test(g, d[spin])
                cg[spin] = cg_test(atoms, g, g_old[spin])
            beta[spin], norm_g[spin] = cg_method(scf, cgform, g[spin], g_old[spin], d_old[spin])
            d[spin] = -atoms.K(g[spin]) + beta[spin] * d_old[spin]
            gt = grad(scf, spin, scf.W + betat * d[spin])
            beta[spin] = betat * dotprod(g[spin], d[spin]) / dotprod(g[spin] - gt, d[spin])
            g_old[spin], d_old[spin] = g[spin], d[spin]

        W_old = np.copy(scf.W)
        scf.W = scf.W + beta * d
        c = cost(scf)
        # If the energy does not go down use the steepest descent step and recalculate the energy
        if c > costs[-1]:
            scf.W = W_old
            for spin in range(atoms.occ.Nspin):
                scf.W[spin] = scf.W[spin] - betat * g[spin]
            c = cost(scf)
            costs.append(c)
            # Do not print cg and linmin if we do the sd step
            if condition(scf, 'sd', costs, norm_g=norm_g):
                break
        else:
            costs.append(c)
            if condition(scf, 'pccg', costs, linmin, cg, norm_g):
                break
    return costs


#: Map minimizer names with their respective implementation.
IMPLEMENTED = {
    'sd': sd,
    'lm': lm,
    'pclm': pclm,
    'cg': cg,
    'pccg': pccg,
    'auto': auto
}



#!/usr/bin/env python3
"""Dispersion correction interface."""
# Somehow I get wrong energies if I try to move these imports into get_Edisp
try:  # noqa: SIM105
    from dftd3.interface import (
        DispersionModel,
        ModifiedRationalDampingParam,
        ModifiedZeroDampingParam,
        OptimizedPowerDampingParam,
        RationalDampingParam,
        ZeroDampingParam,
    )
except ImportError:
    pass
import numpy as np

from ..data import SYMBOL2NUMBER
from ..logger import log


def get_Edisp(scf, version='d3bj', atm=True, xc=None):
    """Calculate the DFT-D3 dispersion correction energy.

    Reference: J. Chem. Phys. 132, 154104.

    Args:
        scf: SCF object.

    Keyword Args:
        version (str): Damping function, can be one of: d3bj', 'd3bjm', 'd3zero', 'd3zerom', 'd3op'.
        atm (bool): Weather to use three-body dispersion energies.
        xc (str | None): Overwrite the functional keyword if the automatic detection does not work.

    Returns:
        float: Dispersion correction energy.
    """
    try:
        import dftd3  # noqa: F401
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[dispersion]".\n\n')
        raise
    # Dictionary of implemented dispersion models
    dispersion_version = {
        'd3bj': RationalDampingParam,
        'd3bjm': ModifiedRationalDampingParam,
        'd3zero': ZeroDampingParam,
        'd3zerom': ModifiedZeroDampingParam,
        'd3op': OptimizedPowerDampingParam
    }

    # Set up input parameters
    # The dispersion correction is only geometry dependent which makes handling the input easy
    atoms = scf.atoms
    positions = atoms.X
    numbers = np.array([SYMBOL2NUMBER[ia] for ia in atoms.atom])
    # Try to determine the method keyword
    if xc is None:
        if scf.xc_type == 'lda':
            method = 'slaterdiracexchange'
            version = 'd3zero'
            log.warning('Dispersion correction for LDA functionals only support d3zero.')
        elif 'pbe_sol' in ''.join(scf.xc):
            method = 'pbesol'
        elif 'pbe' in ''.join(scf.xc):
            method = 'pbe'
        else:
            method = 'pbe'
            log.warning('Functional for the dispersion correction could not be detected, continue'
                        ' with pbe. You may need to overwrite the functional keyword manually.')
    else:
        method = xc

    # Set up the dispersion model object with periodic boundary conditions
    model = DispersionModel(numbers, positions, lattice=atoms.R, periodic=np.array([1, 1, 1]))
    # Calculate the dispersion energy, neglecting the gradient
    res = model.get_dispersion(dispersion_version[version](method=method, atm=atm), grad=False)
    scf.energies.Edisp = res['energy']
    return res['energy']
